from collections import defaultdict
from gettext import gettext as _
from importlib import import_module

from django import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import post_migrate
from django.utils.module_loading import module_has_submodule

from pulpcore.exceptions.plugin import MissingPlugin

VIEWSETS_MODULE_NAME = "viewsets"
SERIALIZERS_MODULE_NAME = "serializers"
URLS_MODULE_NAME = "urls"
MODELRESOURCE_MODULE_NAME = "modelresource"


def pulp_plugin_configs():
    """
    A generator of Pulp plugin AppConfigs

    This makes it easy to iterate over just the installed Pulp plugins when working
    with discovered plugin components.
    """
    for app_config in apps.apps.get_app_configs():
        if isinstance(app_config, PulpPluginAppConfig):
            yield app_config


def get_plugin_config(plugin_app_label):
    """
    A getter of specific pulp plugin config

    This makes it easy to retrieve a config for a specific Pulp plugin when looking for a
    registered model.

    Args:
        plugin_app_label (str): Django app label of the pulp plugin

    Returns:
        :class:`pulpcore.app.apps.PulpPluginAppConfig`: The app config of the Pulp plugin.

    Raises:
        MissingPlugin: When plugin with the requested app label is not installed.
    """
    for config in pulp_plugin_configs():
        if config.label == plugin_app_label:
            return config
    raise MissingPlugin(plugin_app_label)


class PulpPluginAppConfig(apps.AppConfig):
    """AppConfig class. Use this in plugins to identify your app as a Pulp plugin."""

    # Plugin behavior loading should happen in ready(), not in __init__().
    # ready() is called after all models are initialized, and at that point we should
    # be able to safely inspect the plugin modules to look for any components we need
    # to "register" with pulpcore. The viewset registration below is based on Django's
    # own model importing method.

    def __init__(self, app_name, app_module):
        super().__init__(app_name, app_module)

        try:
            self.version
        except AttributeError:
            msg = _(
                "The plugin `{}` is missing a version declaration. Starting with "
                "pulpcore==3.10, plugins are required to define their version on the "
                "PulpPluginAppConfig subclass."
            )
            raise ImproperlyConfigured(msg.format(self.label))

        # Module containing viewsets eg. <module 'pulp_plugin.app.viewsets'
        # from 'pulp_plugin/app/viewsets.py'>. Set by import_viewsets().
        # None if the application doesn't have a viewsets module, automatically set
        # when this app becomes ready.
        self.viewsets_module = None

        # Module containing urlpatterns
        self.urls_module = None

        # Module containing django-import-export ModelResources for a plugin's Models
        self.modelresource_module = None
        # List of classes from self.modelresource_module that can be exported
        self.exportable_classes = None

        # Mapping of model names to viewset lists (viewsets unrelated to models are excluded)
        self.named_viewsets = None
        # Mapping of serializer names to serializers
        self.named_serializers = None

    def ready(self):
        self.import_viewsets()
        self.import_serializers()
        self.import_urls()
        self.import_modelresources()
        post_migrate.connect(
            _populate_access_policies, sender=self, dispatch_uid="my_unique_identifier"
        )

    def import_serializers(self):
        # circular import avoidance
        from pulpcore.app.serializers import ModelSerializer

        self.named_serializers = {}
        if module_has_submodule(self.module, SERIALIZERS_MODULE_NAME):
            # import the serializers module and track any discovered serializers
            serializers_module_name = "{name}.{module}".format(
                name=self.name, module=SERIALIZERS_MODULE_NAME
            )
            self.serializers_module = import_module(serializers_module_name)
            for objname in dir(self.serializers_module):
                obj = getattr(self.serializers_module, objname)
                try:
                    # Any subclass of ModelSerializer that isn't itself ModelSerializer
                    # gets registered in the named_serializers registry.
                    if obj is not ModelSerializer and issubclass(obj, ModelSerializer):
                        self.named_serializers[objname] = obj
                except TypeError:
                    # obj isn't a class, issubclass exploded but obj can be safely filtered out
                    continue

    def import_viewsets(self):
        # TODO do not include imported ViewSets
        # circular import avoidance
        from pulpcore.app.viewsets import NamedModelViewSet

        self.named_viewsets = defaultdict(list)
        if module_has_submodule(self.module, VIEWSETS_MODULE_NAME):
            # import the viewsets module and track any interesting viewsets
            viewsets_module_name = "{name}.{module}".format(
                name=self.name, module=VIEWSETS_MODULE_NAME
            )
            self.viewsets_module = import_module(viewsets_module_name)
            for objname in dir(self.viewsets_module):
                obj = getattr(self.viewsets_module, objname)
                try:
                    # Any subclass of NamedModelViewSet that isn't itself NamedModelViewSet
                    # gets registered in the named_viewsets registry.
                    if obj is not NamedModelViewSet and issubclass(obj, NamedModelViewSet):
                        model = obj.queryset.model
                        self.named_viewsets[model].append(obj)
                except TypeError:
                    # obj isn't a class, issubclass exploded but obj can be safely filtered out
                    continue

    def import_urls(self):
        """
        If a plugin defines a urls.py, include it.
        """
        if module_has_submodule(self.module, URLS_MODULE_NAME) and self.name != "pulpcore.app":
            urls_module_name = "{name}.{module}".format(name=self.name, module=URLS_MODULE_NAME)
            self.urls_module = import_module(urls_module_name)

    def import_modelresources(self):
        """
        If a plugin has a modelresource.py, import it

        (This exists when a plugin knows how to import-export itself)
        """
        if (
            module_has_submodule(self.module, MODELRESOURCE_MODULE_NAME)
            and self.name != "pulpcore.app"
        ):
            modelrsrc_module_name = "{name}.{module}".format(
                name=self.name, module=MODELRESOURCE_MODULE_NAME
            )
            self.modelresource_module = import_module(modelrsrc_module_name)
            self.exportable_classes = self.modelresource_module.IMPORT_ORDER


class PulpAppConfig(PulpPluginAppConfig):
    # The pulpcore app is itself a pulpcore plugin so that it can benefit from
    # the component discovery mechanisms provided by that superclass.

    # The app's importable name
    name = "pulpcore.app"

    # The app label to be used when creating tables, registering models, referencing this app
    # with manage.py, etc. This cannot contain a dot and must not conflict with the name of a
    # package containing a Django app.
    label = "core"

    # The version of this app
    version = "3.16.5"

    def ready(self):
        super().ready()
        from . import checks  # noqa

        post_migrate.connect(_delete_anon_user, sender=self, dispatch_uid="delete_anon_identifier")


def _populate_access_policies(sender, **kwargs):
    from pulpcore.app.util import get_view_urlpattern

    apps = kwargs.get("apps")
    if apps is None:
        from django.apps import apps
    AccessPolicy = apps.get_model("core", "AccessPolicy")
    for viewset_batch in sender.named_viewsets.values():
        for viewset in viewset_batch:
            access_policy = getattr(viewset, "DEFAULT_ACCESS_POLICY", None)
            if access_policy is not None:
                viewset_name = get_view_urlpattern(viewset)
                db_access_policy, created = AccessPolicy.objects.get_or_create(
                    viewset_name=viewset_name, defaults=access_policy
                )
                if created:
                    print(f"Access policy for {viewset_name} created.")
                if not created and not db_access_policy.customized:
                    for key, value in access_policy.items():
                        setattr(db_access_policy, key, value)
                    db_access_policy.save()
                    print(f"Access policy for {viewset_name} updated.")


def _delete_anon_user(sender, **kwargs):
    if settings.ANONYMOUS_USER_NAME is None:
        print(_("Deleting Guardians' AnonymousUser"))
        User = get_user_model()
        try:
            anon = User.objects.get(username="AnonymousUser")
        except User.DoesNotExist:
            pass
        else:
            anon.delete()



import json
import logging
import uuid
from gettext import gettext as _

from django.db import transaction, connection
from django.db.models import Model
from django_guid import get_guid

from pulpcore.app.models import Task
from pulpcore.constants import TASK_STATES
from pulpcore.tasking import util

_logger = logging.getLogger(__name__)


TASK_TIMEOUT = -1  # -1 for infinite timeout


class UUIDEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            return str(obj)
        return super().default(obj)


def _validate_and_get_resources(resources):
    resource_set = set()
    for r in resources:
        if isinstance(r, str):
            resource_set.add(r)
        elif isinstance(r, Model):
            resource_set.add(util.get_url(r))
        else:
            raise ValueError(_("Must be (str|Model)"))
    return list(resource_set)


def dispatch(
    func,
    args=None,
    kwargs=None,
    task_group=None,
    exclusive_resources=None,
    shared_resources=None,
):
    """
    Enqueue a message to Pulp workers with a reservation.

    This method provides normal enqueue functionality, while also requesting necessary locks for
    serialized urls. No two tasks that claim the same resource can execute concurrently. It
    accepts resources which it transforms into a list of urls (one for each resource).

    This method creates a :class:`pulpcore.app.models.Task` object and returns it.

    The values in `args` and `kwargs` must be JSON serializable, but may contain instances of
    ``uuid.UUID``.

    Args:
        func (callable): The function to be run by RQ when the necessary locks are acquired.
        args (tuple): The positional arguments to pass on to the task.
        kwargs (dict): The keyword arguments to pass on to the task.
        task_group (pulpcore.app.models.TaskGroup): A TaskGroup to add the created Task to.
        exclusive_resources (list): A list of resources this task needs exclusive access to while
            running. Each resource can be either a `str` or a `django.models.Model` instance.
        shared_resources (list): A list of resources this task needs non-exclusive access to while
            running. Each resource can be either a `str` or a `django.models.Model` instance.

    Returns (pulpcore.app.models.Task): The Pulp Task that was created.

    Raises:
        ValueError: When `resources` is an unsupported type.
    """

    if exclusive_resources is None:
        exclusive_resources = []

    args_as_json = json.dumps(args, cls=UUIDEncoder)
    kwargs_as_json = json.dumps(kwargs, cls=UUIDEncoder)
    resources = _validate_and_get_resources(exclusive_resources)
    if shared_resources:
        resources.extend(
            (f"shared:{resource}" for resource in _validate_and_get_resources(shared_resources))
        )
    with transaction.atomic():
        task = Task.objects.create(
            state=TASK_STATES.WAITING,
            logging_cid=(get_guid() or ""),
            task_group=task_group,
            name=f"{func.__module__}.{func.__name__}",
            args=args_as_json,
            kwargs=kwargs_as_json,
            parent_task=Task.current(),
            reserved_resources_record=resources,
        )
    # Notify workers
    with connection.connection.cursor() as cursor:
        cursor.execute("NOTIFY pulp_worker_wakeup")
    return task



from gettext import gettext as _

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.core.exceptions import FieldError
from django.db import IntegrityError
from django.shortcuts import get_object_or_404
from django_filters.rest_framework import DjangoFilterBackend
from drf_spectacular.utils import extend_schema
from guardian.models.models import GroupObjectPermission
from rest_framework import mixins, status
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter
from rest_framework.response import Response
from rest_framework.serializers import ValidationError

from pulpcore.app.viewsets import BaseFilterSet, NamedModelViewSet
from pulpcore.app.serializers.user import (
    GroupSerializer,
    GroupUserSerializer,
    PermissionSerializer,
    UserSerializer,
)

User = get_user_model()


class UserFilter(BaseFilterSet):
    """
    FilterSet for User.
    """

    class Meta:
        model = User
        fields = {
            "username": ["exact", "iexact", "in", "contains", "icontains"],
            "first_name": ["exact", "iexact", "in", "contains", "icontains"],
            "last_name": ["exact", "iexact", "in", "contains", "icontains"],
            "email": ["exact", "iexact", "in", "contains", "icontains"],
            "is_active": ["exact"],
            "is_staff": ["exact"],
        }


class UserViewSet(NamedModelViewSet, mixins.RetrieveModelMixin, mixins.ListModelMixin):
    """
    ViewSet for User.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "users"
    filterset_class = UserFilter
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    serializer_class = UserSerializer
    queryset = User.objects.all()
    ordering = ("-date_joined",)

    @extend_schema(description="List user permissions.")
    @action(detail=True, methods=["get"], serializer_class=PermissionSerializer)
    def permissions(self, request, pk):
        """
        List user permissions.
        """
        user = self.get_object()
        object_permission = PermissionSerializer(
            user.userobjectpermission_set.all(), many=True, context={"request": request}
        )
        permissions = PermissionSerializer(user.user_permissions.all(), many=True)
        return Response(object_permission.data + permissions.data)


class GroupFilter(BaseFilterSet):
    """
    FilterSet for Group.
    """

    class Meta:
        model = Group
        fields = {
            "id": ["exact", "in"],
            "name": ["exact", "iexact", "in", "contains", "icontains"],
        }


class GroupViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    """
    ViewSet for Group.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "groups"
    router_lookup = "group"
    filterset_class = GroupFilter
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    serializer_class = GroupSerializer
    queryset = Group.objects.all()
    ordering = ("name",)
    queryset_filtering_required_permission = "auth.view_group"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:auth.add_group",
            },
            {
                "action": ["retrieve"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:auth.view_group",
            },
            {
                "action": ["update", "partial_update"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:auth.change_group",
            },
            {
                "action": ["destroy"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:auth.delete_group",
            },
        ],
        "permissions_assignment": [
            {
                "function": "add_for_object_creator",
                "parameters": None,
                "permissions": [
                    "auth.view_group",
                    "auth.change_group",
                    "auth.delete_group",
                ],
            },
        ],
    }


class GroupModelPermissionViewSet(NamedModelViewSet):
    """
    ViewSet for Model Permissions that belongs to a Group.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "model_permissions"
    nest_prefix = "groups"
    router_lookup = "model_permission"
    parent_viewset = GroupViewSet
    parent_lookup_kwargs = {"group_pk": "group__pk"}
    serializer_class = PermissionSerializer
    queryset = Permission.objects.all()
    pulp_model_alias = "ModelPermission"

    def get_model_permission(self, request):
        """Get model permission"""
        data = {}
        for key, value in request.data.items():
            if key == "permission":
                if "." in value:
                    data["content_type__app_label"], data["codename"] = value.split(".", maxsplit=1)
                else:
                    data["codename"] = value
                continue

            if key == "pulp_href":
                if "id" in data.keys():
                    continue
                data["id"] = value.strip("/").split("/")[-1]
                continue

            data[key] = value

        try:
            permission = Permission.objects.get(**data)
        except (Permission.MultipleObjectsReturned, Permission.DoesNotExist, FieldError) as exc:
            raise ValidationError(str(exc))

        return permission

    @extend_schema(description="Retrieve a model permission from a group.")
    def retrieve(self, request, group_pk, pk):
        instance = get_object_or_404(Permission, pk=pk)
        serializer = PermissionSerializer(instance, context={"group_pk": group_pk})
        return Response(serializer.data)

    @extend_schema(description="List group permissions.", responses={200: PermissionSerializer})
    def list(self, request, group_pk):
        """
        List group model permissions.
        """
        group = Group.objects.get(pk=group_pk)
        queryset = group.permissions.all()

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = PermissionSerializer(page, context={"group_pk": group_pk}, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = PermissionSerializer(queryset, context={"group_pk": group_pk}, many=True)
        return Response(serializer.data)

    @extend_schema(
        description="Add a model permission to a group.", responses={201: PermissionSerializer}
    )
    def create(self, request, group_pk):
        """
        Add a model permission to a group.
        """
        group = Group.objects.get(pk=group_pk)
        permission = self.get_model_permission(request)
        group.permissions.add(permission)
        group.save()
        serializer = PermissionSerializer(permission, context={"group_pk": group_pk})
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    @extend_schema(description="Remove a model permission from a group.")
    def destroy(self, request, group_pk, pk):
        """
        Remove a model permission from a group.
        """
        group = Group.objects.get(pk=group_pk)
        permission = get_object_or_404(Permission, pk=pk)
        group.permissions.remove(permission)
        group.save()
        return Response(status=status.HTTP_204_NO_CONTENT)


class GroupObjectPermissionViewSet(NamedModelViewSet):
    """
    ViewSet for Object Permissions that belongs to a Group.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "object_permissions"
    nest_prefix = "groups"
    router_lookup = "object_permission"
    parent_viewset = GroupViewSet
    parent_lookup_kwargs = {"group_pk": "group__pk"}
    serializer_class = PermissionSerializer
    queryset = Permission.objects.all()
    pulp_model_alias = "ObjectPermission"

    def get_object_pk(self, request):
        """Return an object's pk from the request."""

        if "obj" not in request.data:
            raise ValidationError(_("Please provide 'obj' value"))

        obj_url = request.data["obj"]
        try:
            obj = NamedModelViewSet.get_resource(obj_url)
        except ValidationError:
            raise ValidationError(_("Invalid value for 'obj': {}.").format(obj_url))

        return obj.pk

    def get_model_permission(self, request):
        """Get model permission"""

        if "permission" not in request.data:
            raise ValidationError(_("Please provide 'permission' value"))

        codename = request.data["permission"].split(".")[-1]

        try:
            permission = Permission.objects.get(codename=codename)
        except (Permission.MultipleObjectsReturned, Permission.DoesNotExist, FieldError) as exc:
            raise ValidationError(str(exc))

        return permission

    @extend_schema(description="Retrieve a model permission from a group.")
    def retrieve(self, request, group_pk, pk):
        instance = get_object_or_404(GroupObjectPermission, pk=pk)
        serializer = PermissionSerializer(
            instance, context={"group_pk": group_pk, "request": request}
        )
        return Response(serializer.data)

    @extend_schema(
        description="List group object permissions.", responses={200: PermissionSerializer}
    )
    def list(self, request, group_pk):
        """
        List group object permissions.
        """
        group = Group.objects.get(pk=group_pk)
        queryset = group.groupobjectpermission_set.all()

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = PermissionSerializer(
                page, context={"group_pk": group_pk, "request": request}, many=True
            )
            return self.get_paginated_response(serializer.data)

        serializer = PermissionSerializer(
            queryset, context={"group_pk": group_pk, "request": request}, many=True
        )
        return Response(serializer.data)

    @extend_schema(
        description="Add an object permission to a group.", responses={201: PermissionSerializer}
    )
    def create(self, request, group_pk):
        """
        Create an object permission to a group.
        """
        group = Group.objects.get(pk=group_pk)
        permission = self.get_model_permission(request)
        object_pk = self.get_object_pk(request)
        object_permission = GroupObjectPermission(
            group=group,
            permission=permission,
            content_type_id=permission.content_type_id,
            object_pk=object_pk,
        )

        try:
            object_permission.save()
        except IntegrityError:
            raise ValidationError(_("The assigned permission already exists."))

        serializer = PermissionSerializer(
            object_permission, context={"group_pk": group_pk, "request": request}
        )
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    @extend_schema(description="Remove an object permission from a group.")
    def destroy(self, request, group_pk, pk):
        """
        Delete an object permission from a group.
        """
        object_permission = get_object_or_404(GroupObjectPermission, pk=pk, group_id=group_pk)
        object_permission.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


class GroupUserViewSet(NamedModelViewSet):
    """
    ViewSet for Users that belongs to a Group.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    endpoint_name = "users"
    nest_prefix = "groups"
    router_lookup = "user"
    parent_viewset = GroupViewSet
    parent_lookup_kwargs = {"group_pk": "groups__pk"}
    serializer_class = GroupUserSerializer
    queryset = User.objects.all()

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_group_model_or_obj_perms:auth.view_group",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_group_model_or_obj_perms:auth.change_group",
            },
            {
                "action": ["destroy"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_group_model_or_obj_perms:auth.change_group",
            },
        ],
        "permissions_assignment": [],
    }

    def list(self, request, group_pk):
        """
        List group users.
        """
        group = Group.objects.get(pk=group_pk)
        queryset = group.user_set.all()

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = GroupUserSerializer(page, context={"request": request}, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def create(self, request, group_pk):
        """
        Add a user to a group.
        """
        group = Group.objects.get(pk=group_pk)
        if not request.data:
            raise ValidationError(
                _("Please provide one of the following values for User: 'pk', 'id', 'username'")
            )
        try:
            user = User.objects.get(**request.data)
        except (User.DoesNotExist, FieldError) as exc:
            raise ValidationError(str(exc))
        group.user_set.add(user)
        group.save()
        serializer = GroupUserSerializer(user, context={"request": request})
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def destroy(self, request, group_pk, pk):
        """
        Remove a user from a group.
        """
        group = Group.objects.get(pk=group_pk)
        user = get_object_or_404(User, pk=pk)
        group.user_set.remove(user)
        group.save()
        return Response(status=status.HTTP_204_NO_CONTENT)



# Generated by Django 2.2.24 on 2021-06-15 12:11

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0064_repository_user_hidden'),
        ('core', '0064_add_new_style_task_columns'),
    ]

    operations = [
    ]



# Generated by Django 2.2.20 on 2021-04-27 07:51

import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models

TASK_BATCH_SIZE = 1000


def copy_reserved_resources_record(apps, schema_editor):
    Task = apps.get_model('core', 'Task')

    # Update _reserved_resource_record for all tasks, 1000 tasks at a time.
    # When we hit 1K tasks, go to the db for the batch.
    # Make sure to update the final batch!
    tasks = []
    for task in Task.objects.iterator(chunk_size=TASK_BATCH_SIZE):
        task._reserved_resources_record = list(task.reserved_resources_record.values_list('resource', flat=True))
        tasks.append(task)
        if len(tasks) == TASK_BATCH_SIZE:
            Task.objects.bulk_update(tasks, ["_reserved_resources_record"])
            tasks.clear()

    # Update last set of tasks
    if len(tasks) > 0:
        Task.objects.bulk_update(tasks, ["_reserved_resources_record"])


def purge_reservedresources(apps, schema_editor):
    TaskReservedResource = apps.get_model('core', 'TaskReservedResource')
    TaskReservedResource.objects.all().delete()

    ReservedResource = apps.get_model('core', 'ReservedResource')
    ReservedResource.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0063_repository_retained_versions'),
    ]

    operations = [
        # Purge any ReservedResource entries - if there are any, they're orphans
        migrations.RunPython(
            code=purge_reservedresources,
            reverse_code=migrations.RunPython.noop,
        ),
        # Update entities for the new task-system
        migrations.AddField(
            model_name='task',
            name='args',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
        migrations.AddField(
            model_name='task',
            name='kwargs',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
        migrations.AddField(
            model_name='task',
            name='_reserved_resources_record',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=256), null=True, size=None),
        ),
        migrations.AlterField(
            model_name='task',
            name='_resource_job_id',
            field=models.UUIDField(null=True),
        ),
        migrations.AlterField(
            model_name='progressreport',
            name='state',
            field=models.TextField(choices=[('waiting', 'Waiting'), ('skipped', 'Skipped'), ('running', 'Running'), ('completed', 'Completed'), ('failed', 'Failed'), ('canceled', 'Canceled'), ('canceling', 'Canceling')], default='waiting'),
        ),
        migrations.AlterField(
            model_name='task',
            name='state',
            field=models.TextField(choices=[('waiting', 'Waiting'), ('skipped', 'Skipped'), ('running', 'Running'), ('completed', 'Completed'), ('failed', 'Failed'), ('canceled', 'Canceled'), ('canceling', 'Canceling')]),
        ),
        migrations.AddIndex(
            model_name='task',
            index=models.Index(fields=['pulp_created'], name='core_task_pulp_cr_10223f_idx'),
        ),
        migrations.RunPython(
            code=copy_reserved_resources_record,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.RemoveField(
            model_name='taskreservedresourcerecord',
            name='resource',
        ),
        migrations.RemoveField(
            model_name='taskreservedresourcerecord',
            name='task',
        ),
        migrations.DeleteModel(
            name='ReservedResourceRecord',
        ),
        migrations.DeleteModel(
            name='TaskReservedResourceRecord',
        ),
        migrations.RenameField(
            model_name='task',
            old_name='_reserved_resources_record',
            new_name='reserved_resources_record',
        ),

    ]



"""
Content related Django models.
"""
from gettext import gettext as _

import datetime
import json
import tempfile
import shutil
import subprocess

import gnupg

from functools import lru_cache
from itertools import chain

from django.conf import settings
from django.core import validators
from django.core.files.storage import default_storage
from django.db import IntegrityError, models, transaction
from django.forms.models import model_to_dict
from django.utils.timezone import now
from django_lifecycle import BEFORE_UPDATE, BEFORE_SAVE, hook

from pulpcore.constants import ALL_KNOWN_CONTENT_CHECKSUMS
from pulpcore.app import pulp_hashlib
from pulpcore.app.models import MasterModel, BaseModel, fields, storage
from pulpcore.exceptions import (
    DigestValidationError,
    SizeValidationError,
    MissingDigestValidationError,
    UnsupportedDigestValidationError,
)

# All available digest fields ordered by algorithm strength.
_DIGEST_FIELDS = []
for alg in ("sha512", "sha384", "sha256", "sha224", "sha1", "md5"):
    if alg in settings.ALLOWED_CONTENT_CHECKSUMS:
        _DIGEST_FIELDS.append(alg)

# All available digest fields ordered by relative frequency
# (Better average-case performance in some algorithms with fallback)
_COMMON_DIGEST_FIELDS = []
for alg in ("sha256", "sha512", "sha384", "sha224", "sha1", "md5"):
    if alg in settings.ALLOWED_CONTENT_CHECKSUMS:
        _COMMON_DIGEST_FIELDS.append(alg)

# Available, reliable digest fields ordered by algorithm strength.
_RELIABLE_DIGEST_FIELDS = []
for alg in ("sha512", "sha384", "sha256"):
    if alg in settings.ALLOWED_CONTENT_CHECKSUMS:
        _RELIABLE_DIGEST_FIELDS.append(alg)

# Digest-fields that are NOT ALLOWED
_FORBIDDEN_DIGESTS = set(ALL_KNOWN_CONTENT_CHECKSUMS).difference(settings.ALLOWED_CONTENT_CHECKSUMS)


class BulkCreateManager(models.Manager):
    """
    A manager that provides a bulk_get_or_create()
    """

    def bulk_get_or_create(self, objs, batch_size=None):
        """
        Insert the list of objects into the database and get existing objects from the database.

        Do *not* call save() on each of the instances, do not send any pre/post_save signals,
        and do not set the primary key attribute if it is an autoincrement field (except if
        features.can_return_ids_from_bulk_insert=True). Multi-table models are not supported.

        If an IntegrityError is raised while performing a bulk insert, this method falls back to
        inserting each instance individually. The already-existing instances are retrieved from
        the database and returned with the other newly created instances.

        Args:
            objs (iterable of models.Model): an iterable of Django Model instances
            batch_size (int): how many are created in a single query

        Returns:
            List of instances that were inserted into the database.
        """

        objs = list(objs)
        try:
            with transaction.atomic():
                return super().bulk_create(objs, batch_size=batch_size)
        except IntegrityError:
            for i in range(len(objs)):
                try:
                    with transaction.atomic():
                        objs[i].save()
                except IntegrityError:
                    objs[i] = objs[i].__class__.objects.get(objs[i].q())
        return objs


class BulkTouchQuerySet(models.QuerySet):
    """
    A query set that provides ``touch()``.
    """

    def touch(self):
        """
        Update the ``timestamp_of_interest`` on all objects of the query.

        Postgres' UPDATE call doesn't support order-by. This can (and does) result in deadlocks in
        high-concurrency environments, when using touch() on overlapping data sets. In order to
        prevent this, we choose to SELECT FOR UPDATE with SKIP LOCKS == True, and only update
        the rows that we were able to get locks on. Since a previously-locked-row implies
        that updating that row's timestamp-of-interest is the responsibility of whoever currently
        owns it, this results in correct data, while closing the window on deadlocks.
        """
        with transaction.atomic():
            sub_q = self.order_by("pk").select_for_update(skip_locked=True)
            return self.filter(pk__in=sub_q).update(timestamp_of_interest=now())


class QueryMixin:
    """
    A mixin that provides models with querying utilities.
    """

    def q(self):
        """
        Returns a Q object that represents the model
        """
        if not self._state.adding:
            return models.Q(pk=self.pk)
        try:
            kwargs = self.natural_key_dict()
        except AttributeError:
            all_kwargs = model_to_dict(self)
            kwargs = {k: v for k, v in all_kwargs.items() if v}
        return models.Q(**kwargs)


class HandleTempFilesMixin:
    """
    A mixin that provides methods for handling temporary files.
    """

    def save(self, *args, **kwargs):
        """
        Saves Model and closes the file associated with the Model

        Args:
            args (list): list of positional arguments for Model.save()
            kwargs (dict): dictionary of keyword arguments to pass to Model.save()
        """
        try:
            super().save(*args, **kwargs)
        finally:
            self.file.close()

    def delete(self, *args, **kwargs):
        """
        Deletes Model and the file associated with the Model

        Args:
            args (list): list of positional arguments for Model.delete()
            kwargs (dict): dictionary of keyword arguments to pass to Model.delete()
        """
        super().delete(*args, **kwargs)
        self.file.delete(save=False)


class ArtifactManager(BulkCreateManager):
    def orphaned(self, orphan_protection_time):
        """Returns set of orphaned artifacts that are ready to be cleaned up."""
        expiration = now() - datetime.timedelta(minutes=orphan_protection_time)
        return self.filter(content_memberships__isnull=True, timestamp_of_interest__lt=expiration)


class Artifact(HandleTempFilesMixin, BaseModel):
    """
    A file associated with a piece of content.

    When calling `save()` on an Artifact, if the file is not stored in Django's storage backend, it
    is moved into place then.

    Artifact is compatible with Django's `bulk_create()` method.

    Fields:

        file (pulpcore.app.models.fields.ArtifactFileField): The stored file. This field should
            be set using an absolute path to a temporary file.
            It also accepts `class:django.core.files.File`.
        size (models.BigIntegerField): The size of the file in bytes.
        md5 (models.CharField): The MD5 checksum of the file.
        sha1 (models.CharField): The SHA-1 checksum of the file.
        sha224 (models.CharField): The SHA-224 checksum of the file.
        sha256 (models.CharField): The SHA-256 checksum of the file (REQUIRED).
        sha384 (models.CharField): The SHA-384 checksum of the file.
        sha512 (models.CharField): The SHA-512 checksum of the file.
        timestamp_of_interest (models.DateTimeField): timestamp that prevents orphan cleanup
    """

    def storage_path(self, name):
        """
        Callable used by FileField to determine where the uploaded file should be stored.

        Args:
            name (str): Original name of uploaded file. It is ignored by this method because the
                sha256 checksum is used to determine a file path instead.
        """
        return storage.get_artifact_path(self.sha256)

    file = fields.ArtifactFileField(null=False, upload_to=storage_path, max_length=255)
    size = models.BigIntegerField(null=False)
    md5 = models.CharField(max_length=32, null=True, unique=False, db_index=True)
    sha1 = models.CharField(max_length=40, null=True, unique=False, db_index=True)
    sha224 = models.CharField(max_length=56, null=True, unique=False, db_index=True)
    sha256 = models.CharField(max_length=64, null=False, unique=True, db_index=True)
    sha384 = models.CharField(max_length=96, null=True, unique=True, db_index=True)
    sha512 = models.CharField(max_length=128, null=True, unique=True, db_index=True)
    timestamp_of_interest = models.DateTimeField(auto_now=True)

    objects = ArtifactManager.from_queryset(BulkTouchQuerySet)()

    # All available digest fields ordered by algorithm strength.
    DIGEST_FIELDS = _DIGEST_FIELDS

    # All available digest fields ordered by relative frequency
    # (Better average-case performance in some algorithms with fallback)
    COMMON_DIGEST_FIELDS = _COMMON_DIGEST_FIELDS

    # Available, reliable digest fields ordered by algorithm strength.
    RELIABLE_DIGEST_FIELDS = _RELIABLE_DIGEST_FIELDS

    # Digest-fields that are NOT ALLOWED
    FORBIDDEN_DIGESTS = _FORBIDDEN_DIGESTS

    @hook(BEFORE_SAVE)
    def before_save(self):
        """
        Pre-save hook that validates checksum rules prior to allowing an Artifact to be saved.

        An Artifact with any checksums from the FORBIDDEN set will fail to save while raising
        an UnsupportedDigestValidationError exception.

        Similarly, any checksums in DIGEST_FIELDS that is NOT set will raise a
        MissingDigestValidationError exception.

        Raises:
            :class: `~pulpcore.exceptions.UnsupportedDigestValidationError`: When any of the
                keys on FORBIDDEN_DIGESTS are set for the Artifact
            :class: `~pulpcore.exceptions.MissingDigestValidationError`: When any of the
                keys on DIGEST_FIELDS are found to be missing from the Artifact
        """
        bad_keys = [k for k in self.FORBIDDEN_DIGESTS if getattr(self, k)]
        if bad_keys:
            raise UnsupportedDigestValidationError(
                _("Checksum algorithms {} are forbidden for this Pulp instance.").format(bad_keys)
            )

        missing_keys = [k for k in self.DIGEST_FIELDS if not getattr(self, k)]
        if missing_keys:
            raise MissingDigestValidationError(
                _("Missing required checksum algorithms {}.").format(missing_keys)
            )

    def q(self):
        if not self._state.adding:
            return models.Q(pk=self.pk)
        for digest_name in self.DIGEST_FIELDS:
            digest_value = getattr(self, digest_name)
            if digest_value:
                return models.Q(**{digest_name: digest_value})
        return models.Q()

    def is_equal(self, other):
        """
        Is equal by matching digest.

        Args:
            other (pulpcore.app.models.Artifact): A artifact to match.

        Returns:
            bool: True when equal.
        """
        for field in self.RELIABLE_DIGEST_FIELDS:
            digest = getattr(self, field)
            if not digest:
                continue
            if digest == getattr(other, field):
                return True
        return False

    @staticmethod
    def init_and_validate(file, expected_digests=None, expected_size=None):
        """
        Initialize an in-memory Artifact from a file, and validate digest and size info.

        This accepts both a path to a file on-disk or a
        :class:`~pulpcore.app.files.PulpTemporaryUploadedFile`.

        Args:
            file (:class:`~pulpcore.app.files.PulpTemporaryUploadedFile` or str): The
                PulpTemporaryUploadedFile to create the Artifact from or a string with the full path
                to the file on disk.
            expected_digests (dict): Keyed on the algorithm name provided by hashlib and stores the
                value of the expected digest. e.g. {'md5': '912ec803b2ce49e4a541068d495ab570'}
            expected_size (int): The number of bytes the download is expected to have.

        Raises:
            :class:`~pulpcore.exceptions.DigestValidationError`: When any of the ``expected_digest``
                values don't match the digest of the data
            :class:`~pulpcore.exceptions.SizeValidationError`: When the ``expected_size`` value
                doesn't match the size of the data
            :class:`~pulpcore.exceptions.UnsupportedDigestValidationError`: When any of the
                ``expected_digest`` algorithms aren't in the ALLOWED_CONTENT_CHECKSUMS list
        Returns:
            An in-memory, unsaved :class:`~pulpcore.plugin.models.Artifact`
        """
        if isinstance(file, str):
            with open(file, "rb") as f:
                hashers = {n: pulp_hashlib.new(n) for n in Artifact.DIGEST_FIELDS}
                size = 0
                while True:
                    chunk = f.read(1048576)  # 1 megabyte
                    if not chunk:
                        break
                    for algorithm in hashers.values():
                        algorithm.update(chunk)
                    size = size + len(chunk)
        else:
            size = file.size
            hashers = file.hashers

        if expected_size:
            if size != expected_size:
                raise SizeValidationError()

        if expected_digests:
            for algorithm, expected_digest in expected_digests.items():
                if algorithm not in hashers:
                    raise UnsupportedDigestValidationError(
                        _("Checksum algorithm {} forbidden for this Pulp instance.").format(
                            algorithm
                        )
                    )
                if expected_digest != hashers[algorithm].hexdigest():
                    raise DigestValidationError()

        attributes = {"size": size, "file": file}
        for algorithm in Artifact.DIGEST_FIELDS:
            attributes[algorithm] = hashers[algorithm].hexdigest()

        return Artifact(**attributes)

    @classmethod
    def from_pulp_temporary_file(cls, temp_file):
        """
        Creates an Artifact from PulpTemporaryFile.

        Returns:
            An saved :class:`~pulpcore.plugin.models.Artifact`
        """
        artifact_file = default_storage.open(temp_file.file.name)
        with tempfile.NamedTemporaryFile("wb") as new_file:
            shutil.copyfileobj(artifact_file, new_file)
            new_file.flush()
            artifact = cls.init_and_validate(new_file.name)
            artifact.save()
        temp_file.delete()
        return artifact

    def touch(self):
        """Update timestamp_of_interest."""
        self.save(update_fields=["timestamp_of_interest"])


class PulpTemporaryFile(HandleTempFilesMixin, BaseModel):
    """
    A temporary file saved to the storage backend.

    Commonly used to pass data to one or more tasks.

    Fields:

        file (pulpcore.app.models.fields.ArtifactFileField): The stored file. This field should
            be set using an absolute path to a temporary file.
            It also accepts `class:django.core.files.File`.
    """

    def storage_path(self, name):
        """
        Callable used by FileField to determine where the uploaded file should be stored.

        Args:
            name (str): Original name of uploaded file. It is ignored by this method because the
                pulp_id is used to determine a file path instead.
        """
        return storage.get_temp_file_path(self.pulp_id)

    file = fields.ArtifactFileField(null=False, upload_to=storage_path, max_length=255)

    @staticmethod
    def init_and_validate(file, expected_digests=None, expected_size=None):
        """
        Initialize an in-memory PulpTemporaryFile from a file, and validate digest and size info.

        This accepts both a path to a file on-disk or a
        :class:`~pulpcore.app.files.PulpTemporaryUploadedFile`.

        Args:
            file (:class:`~pulpcore.app.files.PulpTemporaryUploadedFile` or str): The
                PulpTemporaryUploadedFile to create the PulpTemporaryFile from or a string with the
                full path to the file on disk.
            expected_digests (dict): Keyed on the algorithm name provided by hashlib and stores the
                value of the expected digest. e.g. {'md5': '912ec803b2ce49e4a541068d495ab570'}
            expected_size (int): The number of bytes the download is expected to have.

        Raises:
            :class:`~pulpcore.exceptions.DigestValidationError`: When any of the ``expected_digest``
                values don't match the digest of the data
            :class:`~pulpcore.exceptions.SizeValidationError`: When the ``expected_size`` value
                doesn't match the size of the data
            :class:`~pulpcore.exceptions.UnsupportedDigestValidationError`: When any of the
                ``expected_digest`` algorithms aren't in the ALLOWED_CONTENT_CHECKSUMS list

        Returns:
            An in-memory, unsaved :class:`~pulpcore.plugin.models.PulpTemporaryFile`
        """
        if not expected_digests and not expected_size:
            return PulpTemporaryFile(file=file)

        if isinstance(file, str):
            with open(file, "rb") as f:
                hashers = {n: pulp_hashlib.new(n) for n in expected_digests.keys()}
                size = 0
                while True:
                    chunk = f.read(1048576)  # 1 megabyte
                    if not chunk:
                        break
                    for algorithm in hashers.values():
                        algorithm.update(chunk)
                    size = size + len(chunk)
        else:
            size = file.size
            hashers = file.hashers

        if expected_size:
            if size != expected_size:
                raise SizeValidationError()

        if expected_digests:
            for algorithm, expected_digest in expected_digests.items():
                if algorithm not in hashers:
                    raise UnsupportedDigestValidationError(
                        _("Checksum algorithm {} forbidden for this Pulp instance.").format(
                            algorithm
                        )
                    )
                if expected_digest != hashers[algorithm].hexdigest():
                    raise DigestValidationError()

        return PulpTemporaryFile(file=file)


class ContentManager(BulkCreateManager):
    def orphaned(self, orphan_protection_time, content_pks=None):
        """Returns set of orphaned content that is ready to be cleaned up."""
        expiration = now() - datetime.timedelta(minutes=orphan_protection_time)
        if content_pks:
            return self.filter(
                version_memberships__isnull=True,
                timestamp_of_interest__lt=expiration,
                pk__in=content_pks,
            )
        return self.filter(version_memberships__isnull=True, timestamp_of_interest__lt=expiration)


ContentManager = ContentManager.from_queryset(BulkTouchQuerySet)


class Content(MasterModel, QueryMixin):
    """
    A piece of managed content.

    Fields:
        upstream_id (models.UUIDField) : identifier of content imported from an 'upstream' Pulp
        timestamp_of_interest (models.DateTimeField): timestamp that prevents orphan cleanup

    Relations:

        _artifacts (models.ManyToManyField): Artifacts related to Content through ContentArtifact
    """

    PROTECTED_FROM_RECLAIM = True

    TYPE = "content"
    repo_key_fields = ()  # Used by pulpcore.plugin.repo_version_utils.remove_duplicates
    upstream_id = models.UUIDField(null=True)  # Used by PulpImport/Export processing

    _artifacts = models.ManyToManyField(Artifact, through="ContentArtifact")
    timestamp_of_interest = models.DateTimeField(auto_now=True)

    objects = ContentManager()

    class Meta:
        verbose_name_plural = "content"
        unique_together = ()

    @classmethod
    def natural_key_fields(cls):
        """
        Returns a tuple of the natural key fields which usually equates to unique_together fields.

        This can be overwritten in subclasses and should return a tuple of field names.
        """
        return tuple(chain.from_iterable(cls._meta.unique_together))

    @classmethod
    @lru_cache(typed=True)
    def _sanitized_natural_key_fields(cls):
        """
        This function translates the names of the key fields to their attname.

        In case of foreign keys, this decodes to the corresponding `<...>_id` field preventing
        extra DB accesses to fetch the related objects.
        """
        return tuple(getattr(cls, field).field.attname for field in cls.natural_key_fields())

    def natural_key(self):
        """
        Get the model's natural key based on natural_key_fields.

        Returns:
            tuple: The natural key.
        """
        return tuple(getattr(self, f) for f in self._sanitized_natural_key_fields())

    def natural_key_dict(self):
        """
        Get the model's natural key as a dictionary of keys and values.
        """
        return {f: getattr(self, f) for f in self._sanitized_natural_key_fields()}

    @staticmethod
    def init_from_artifact_and_relative_path(artifact, relative_path):
        """
        Return an instance of the specific content by inspecting an artifact.

        Plugin writers are expected to override this method with an implementation for a specific
        content type.

        For example::

            if path.isabs(relative_path):
                raise ValueError(_("Relative path can't start with '/'."))
            return FileContent(relative_path=relative_path, digest=artifact.sha256)

        Args:
            artifact (:class:`~pulpcore.plugin.models.Artifact`): An instance of an Artifact
            relative_path (str): Relative path for the content

        Raises:
            ValueError: If relative_path starts with a '/'.

        Returns:
            An un-saved instance of :class:`~pulpcore.plugin.models.Content` sub-class.
        """
        raise NotImplementedError()

    def touch(self):
        """Update timestamp_of_interest."""
        self.save(update_fields=["timestamp_of_interest"])


class ContentArtifact(BaseModel, QueryMixin):
    """
    A relationship between a Content and an Artifact.

    Serves as a through model for the '_artifacts' ManyToManyField in Content.
    Artifact is protected from deletion if it's present in a ContentArtifact relationship.
    """

    artifact = models.ForeignKey(
        Artifact, on_delete=models.PROTECT, null=True, related_name="content_memberships"
    )
    content = models.ForeignKey(Content, on_delete=models.CASCADE)
    relative_path = models.TextField()

    objects = BulkCreateManager()

    class Meta:
        unique_together = ("content", "relative_path")

    @staticmethod
    def sort_key(ca):
        """
        Static method for defining a sort-key for a specified ContentArtifact.

        Sorting lists of ContentArtifacts is critical for avoiding deadlocks in high-concurrency
        environments, when multiple workers may be operating on similar sets of content at the
        same time. Providing a stable sort-order becomes problematic when the CAs in question
        haven't been persisted - in that case, pulp_id can't be relied on, as it will change
        when the object is stored in the DB and its "real" key is generated.

        This method produces a key based on the content/artifact represented by the CA.

        Args:
            ca (:class:`~pulpcore.plugin.models.ContentArtifact`): The CA we need a key for

        Returns:
            a tuple of (str(content-key), str(artifact-key)) that can be reliably sorted on
        """
        c_key = ""
        a_key = ""
        # It's possible to only have one of content/artifact - handle that
        if ca.content:
            # Some key-fields aren't str, handle that
            c_key = "".join(map(str, ca.content.natural_key()))
        if ca.artifact:
            a_key = str(ca.artifact.sha256)
        return c_key, a_key


class RemoteArtifactQuerySet(models.QuerySet):
    """QuerySet that provides methods for querying RemoteArtifact."""

    def acs(self):
        """Return RemoteArtifacts if they belong to an ACS."""
        return self.filter(remote__alternatecontentsource__isnull=False)

    def order_by_acs(self):
        """Order RemoteArtifacts returning ones with ACSes first."""
        return self.annotate(acs_count=models.Count("remote__alternatecontentsource")).order_by(
            "-acs_count"
        )


class RemoteArtifact(BaseModel, QueryMixin):
    """
    Represents a content artifact that is provided by a remote (external) repository.

    Remotes that want to support deferred download policies should use this model to store
    information required for downloading an Artifact at some point in the future. At a minimum this
    includes the URL, the ContentArtifact, and the Remote that created it. It can also store
    expected size and any expected checksums.

    Fields:

        url (models.TextField): The URL where the artifact can be retrieved.
        size (models.BigIntegerField): The expected size of the file in bytes.
        md5 (models.CharField): The expected MD5 checksum of the file.
        sha1 (models.CharField): The expected SHA-1 checksum of the file.
        sha224 (models.CharField): The expected SHA-224 checksum of the file.
        sha256 (models.CharField): The expected SHA-256 checksum of the file.
        sha384 (models.CharField): The expected SHA-384 checksum of the file.
        sha512 (models.CharField): The expected SHA-512 checksum of the file.

    Relations:

        content_artifact (:class:`pulpcore.app.models.ForeignKey`):
            ContentArtifact associated with this RemoteArtifact.
        remote (:class:`django.db.models.ForeignKey`): Remote that created the
            RemoteArtifact.
    """

    url = models.TextField(validators=[validators.URLValidator])
    size = models.BigIntegerField(null=True)
    md5 = models.CharField(max_length=32, null=True)
    sha1 = models.CharField(max_length=40, null=True)
    sha224 = models.CharField(max_length=56, null=True)
    sha256 = models.CharField(max_length=64, null=True)
    sha384 = models.CharField(max_length=96, null=True)
    sha512 = models.CharField(max_length=128, null=True)

    content_artifact = models.ForeignKey(ContentArtifact, on_delete=models.CASCADE)
    remote = models.ForeignKey("Remote", on_delete=models.CASCADE)

    objects = BulkCreateManager.from_queryset(RemoteArtifactQuerySet)()

    def validate_checksums(self):
        """Validate if RemoteArtifact has allowed checksum or no checksum at all."""
        if not any(
            [
                checksum_type
                for checksum_type in ALL_KNOWN_CONTENT_CHECKSUMS
                if getattr(self, checksum_type, False)
            ]
        ):
            return
        if not any(
            [
                checksum_type
                for checksum_type in Artifact.DIGEST_FIELDS
                if getattr(self, checksum_type, False)
            ]
        ):
            raise UnsupportedDigestValidationError(
                _(
                    "On-demand content located at the url {} contains forbidden checksum type,"
                    "thus cannot be synced."
                    "You can allow checksum type with 'ALLOWED_CONTENT_CHECKSUMS' setting."
                ).format(self.url)
            )

    class Meta:
        unique_together = ("content_artifact", "remote")


class SigningService(BaseModel):
    """
    A model used for producing signatures.

    Fields:
        name (models.TextField):
            A unique name describing a script (or executable) used for signing.
        public_key (models.TextField):
            The value of the public key.
        script (models.TextField):
            An absolute path to an external signing script (or executable).

    """

    name = models.TextField(db_index=True, unique=True)
    public_key = models.TextField()
    pubkey_fingerprint = models.TextField()
    script = models.TextField()

    def sign(self, filename):
        """
        Signs the file provided via 'filename' by invoking an external script (or executable).

        The external script is run as a subprocess. This is done in the expectation that the script
        has been validated as an external signing service by the validate() method. This validation
        is currently only done when creating the signing service object, but not at the time of use.

        Args:
            filename (str): A relative path to a file which is intended to be signed.

        Raises:
            RuntimeError: If the return code of the script is not equal to 0.

        Returns:
            A dictionary as validated by the validate() method.
        """
        completed_process = subprocess.run(
            [self.script, filename], env={}, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        if completed_process.returncode != 0:
            raise RuntimeError(str(completed_process.stderr))

        try:
            return_value = json.loads(completed_process.stdout)
        except json.JSONDecodeError:
            raise RuntimeError("The signing service script did not return valid JSON!")

        return return_value

    def validate(self):
        """
        Ensure that the external signing script produces the desired behaviour.

        With desired behaviour we mean the behaviour as validated by this method. Subclasses are
        required to implement this method. Works by calling the sign() method on some test data, and
        verifying that any desired signatures/signature files are returned and are valid. Must also
        enforce the desired return value format of the sign() method.

        Raises:
            RuntimeError: If the script failed to produce the desired outcome for the test data.
        """
        raise NotImplementedError("Subclasses must implement a validate() method.")

    def save(self, *args, **kwargs):
        """
        Save a signing service to the database (unless it fails to validate).
        """
        if not self.public_key:
            raise RuntimeError(
                _(
                    "The public key needs to be specified during the new instance creation. Other "
                    "ways of providing the public key are considered deprecated as of the release "
                    "(3.10)."
                )
            )

        self.validate()
        super().save(*args, **kwargs)

    @hook(BEFORE_UPDATE)
    def on_update(self):
        raise RuntimeError(
            _(
                "The signing service is immutable. It is advised to create a new signing service "
                "when a change is required."
            )
        )


class AsciiArmoredDetachedSigningService(SigningService):
    """
    A model used for creating detached ASCII armored signatures.
    """

    def validate(self):
        """
        Validate a signing service for a detached ASCII armored signature.

        The validation seeks to ensure that the sign() method returns a dict as follows:

        {"file": "signed_file.xml", "signature": "signed_file.asc"}

        The method creates a file with some content, signs the file, and checks if the
        signature can be verified by the provided public key.

        Raises:
            RuntimeError: If the validation has failed.
        """
        gpg = gnupg.GPG()
        with tempfile.TemporaryDirectory() as temp_directory_name:
            with tempfile.NamedTemporaryFile(dir=temp_directory_name) as temp_file:
                temp_file.write(b"arbitrary data")
                temp_file.flush()
                signed = self.sign(temp_file.name)

                with open(signed["signature"], "rb") as fp:
                    verified = gpg.verify_file(fp, temp_file.name)
                    if verified.trust_level is None or verified.trust_level < verified.TRUST_FULLY:
                        raise RuntimeError(
                            _(
                                "The signature could not be verified or the trust level is too "
                                "low. The signing script may generate invalid signatures."
                            )
                        )
                    elif verified.pubkey_fingerprint != self.pubkey_fingerprint:
                        raise RuntimeError(
                            _(
                                "Fingerprints of the provided public key and the verified public "
                                "key are not equal. The signing script is probably not valid."
                            )
                        )



# Generated by Django 2.2.7 on 2019-11-25 19:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0016_charfield_to_textfield'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='task',
            name='parent',
        ),
    ]



# Generated by Django 2.2.17 on 2021-01-20 14:34

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0052_tasking_logging_cid'),
    ]

    operations = [
        migrations.AddField(
            model_name='remote',
            name='headers',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True),
        ),
    ]



from pulpcore.app.models import (
    Artifact,
    Content,
    ContentArtifact,
    ProgressReport,
    PublishedMetadata,
    Repository,
    RepositoryVersion,
)


def reclaim_space(repo_pks, keeplist_rv_pks=None, force=False):
    """
    This task frees-up disk space by removing Artifact files from the filesystem for Content
    exclusive to the list of provided repos.

    Note: content marked as `proctected` will be excluded from the reclaim disk space.

    Kwargs:
        repo_pks (list): A list of repo pks the disk reclaim space is performed on.
        keeplist_rv_pks (list): A list of repo version pks that will be excluded from the reclaim
        disk space.
        force (bool): If True, uploaded content will be taken into account.

    """
    rest_of_repos = Repository.objects.exclude(pk__in=repo_pks)
    c_keep_qs = Content.objects.filter(repositories__in=rest_of_repos)
    c_reclaim_qs = Content.objects.filter(repositories__in=repo_pks)
    c_reclaim_qs = c_reclaim_qs.exclude(
        pk__in=c_keep_qs, pulp_type=PublishedMetadata.get_pulp_type()
    )

    if keeplist_rv_pks:
        rv_qs = RepositoryVersion.objects.filter(pk__in=keeplist_rv_pks)
        rv_content = Content.objects.none()
        for rv in rv_qs.iterator():
            rv_content |= rv.content
        c_reclaim_qs = c_reclaim_qs.exclude(pk__in=rv_content)

    content_distinct = c_reclaim_qs.distinct("pulp_type")
    unprotected = []
    for content in content_distinct:
        if not content.cast().PROTECTED_FROM_RECLAIM:
            unprotected.append(content.pulp_type)

    ca_qs = ContentArtifact.objects.filter(
        content__in=c_reclaim_qs.values("pk"), artifact__isnull=False
    )
    if not force:
        ca_qs = ca_qs.filter(remoteartifact__isnull=False)
    artifact_pks = set()
    ca_to_update = []
    for ca in ca_qs.iterator():
        if ca.content.pulp_type in unprotected:
            artifact_pks.add(ca.artifact.pk)
            ca.artifact = None
            ca_to_update.append(ca)

    ContentArtifact.objects.bulk_update(objs=ca_to_update, fields=["artifact"], batch_size=1000)
    artifacts_to_delete = Artifact.objects.filter(pk__in=artifact_pks)
    progress_bar = ProgressReport(
        message="Reclaim disk space",
        total=artifacts_to_delete.count(),
        code="reclaim-space.artifact",
        done=0,
        state="running",
    )
    progress_bar.save()

    counter = 0
    interval = 100
    for artifact in artifacts_to_delete.iterator():
        # we need to manually call delete() because it cleans up the file on the filesystem
        artifact.delete()
        progress_bar.done += 1
        counter += 1

        if counter >= interval:
            progress_bar.save()
            counter = 0

    progress_bar.state = "completed"
    progress_bar.save()



from collections import defaultdict

from asgiref.sync import sync_to_async
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, transaction
from django.db.models import Q

from pulpcore.plugin.sync import sync_to_async_iterable

from pulpcore.plugin.models import Content, ContentArtifact, ProgressReport

from .api import Stage


class QueryExistingContents(Stage):
    """
    A Stages API stage that saves :attr:`DeclarativeContent.content` objects and saves its related
    :class:`~pulpcore.plugin.models.ContentArtifact` objects too.

    This stage expects :class:`~pulpcore.plugin.stages.DeclarativeContent` units from `self._in_q`
    and inspects their associated :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects. Each
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` object stores one
    :class:`~pulpcore.plugin.models.Artifact`.

    This stage inspects any "unsaved" Content unit objects and searches for existing saved Content
    units inside Pulp with the same unit key. Any existing Content objects found, replace their
    "unsaved" counterpart in the :class:`~pulpcore.plugin.stages.DeclarativeContent` object.

    Each :class:`~pulpcore.plugin.stages.DeclarativeContent` is sent to `self._out_q` after it has
    been handled.

    This stage drains all available items from `self._in_q` and batches everything into one large
    call to the db for efficiency.
    """

    async def run(self):
        """
        The coroutine for this stage.

        Returns:
            The coroutine for this stage.
        """
        async for batch in self.batches():
            content_q_by_type = defaultdict(lambda: Q(pk__in=[]))
            d_content_by_nat_key = defaultdict(list)
            for d_content in batch:
                if d_content.content._state.adding:
                    model_type = type(d_content.content)
                    unit_q = d_content.content.q()
                    content_q_by_type[model_type] = content_q_by_type[model_type] | unit_q
                    d_content_by_nat_key[d_content.content.natural_key()].append(d_content)

            for model_type, content_q in content_q_by_type.items():
                try:
                    await sync_to_async(model_type.objects.filter(content_q).touch)()
                except AttributeError:
                    raise TypeError(
                        "Plugins which declare custom ORM managers on their content classes "
                        "should have those managers inherit from "
                        "pulpcore.plugin.models.ContentManager."
                    )
                async for result in sync_to_async_iterable(
                    model_type.objects.filter(content_q).iterator()
                ):
                    for d_content in d_content_by_nat_key[result.natural_key()]:
                        d_content.content = result

            for d_content in batch:
                await self.put(d_content)


class ContentSaver(Stage):
    """
    A Stages API stage that saves :attr:`DeclarativeContent.content` objects and saves its related
    :class:`~pulpcore.plugin.models.ContentArtifact` objects too.

    This stage expects :class:`~pulpcore.plugin.stages.DeclarativeContent` units from `self._in_q`
    and inspects their associated :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects. Each
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` object stores one
    :class:`~pulpcore.plugin.models.Artifact`.

    Each "unsaved" Content objects is saved and a :class:`~pulpcore.plugin.models.ContentArtifact`
    objects too.

    Each :class:`~pulpcore.plugin.stages.DeclarativeContent` is sent to after it has been handled.

    This stage drains all available items from `self._in_q` and batches everything into one large
    call to the db for efficiency.
    """

    async def run(self):
        """
        The coroutine for this stage.

        Returns:
            The coroutine for this stage.
        """
        async for batch in self.batches():

            def process_batch():
                content_artifact_bulk = []
                to_update_ca_query = ContentArtifact.objects.none()
                to_update_ca_bulk = []
                to_update_ca_artifact = {}
                with transaction.atomic():
                    self._pre_save(batch)
                    # Process the batch in dc.content.natural_keys order.
                    # This prevents deadlocks when we're processing the same/similar content
                    # in concurrent workers.
                    batch.sort(key=lambda x: "".join(map(str, x.content.natural_key())))
                    for d_content in batch:
                        # Are we saving to the database for the first time?
                        content_already_saved = not d_content.content._state.adding
                        if not content_already_saved:
                            try:
                                with transaction.atomic():
                                    d_content.content.save()
                            except IntegrityError as e:
                                try:
                                    d_content.content = d_content.content.__class__.objects.get(
                                        d_content.content.q()
                                    )
                                except ObjectDoesNotExist:
                                    raise e
                            else:
                                for d_artifact in d_content.d_artifacts:
                                    if not d_artifact.artifact._state.adding:
                                        artifact = d_artifact.artifact
                                    else:
                                        # set to None for on-demand synced artifacts
                                        artifact = None
                                    content_artifact = ContentArtifact(
                                        content=d_content.content,
                                        artifact=artifact,
                                        relative_path=d_artifact.relative_path,
                                    )
                                    content_artifact_bulk.append(content_artifact)
                                continue
                        # When the Content already exists, check if ContentArtifacts need to be
                        # updated
                        for d_artifact in d_content.d_artifacts:
                            if not d_artifact.artifact._state.adding:
                                # the artifact is already present in the database; update references
                                # Creating one large query and one large dictionary
                                to_update_ca_query |= ContentArtifact.objects.filter(
                                    content=d_content.content,
                                    relative_path=d_artifact.relative_path,
                                )
                                key = (d_content.content.pk, d_artifact.relative_path)
                                to_update_ca_artifact[key] = d_artifact.artifact
                    # Query db once and update each object in memory for bulk_update call
                    for content_artifact in to_update_ca_query.iterator():
                        key = (content_artifact.content_id, content_artifact.relative_path)
                        # Maybe remove dict elements after to reduce memory?
                        content_artifact.artifact = to_update_ca_artifact[key]
                        to_update_ca_bulk.append(content_artifact)

                    # Sort the lists we're about to do bulk updates/creates on.
                    # We know to_update_ca_bulk entries already are in the DB, so we can enforce
                    # order just using pulp_id.
                    to_update_ca_bulk.sort(key=lambda x: x.pulp_id)
                    content_artifact_bulk.sort(key=lambda x: ContentArtifact.sort_key(x))

                    ContentArtifact.objects.bulk_update(to_update_ca_bulk, ["artifact"])
                    ContentArtifact.objects.bulk_get_or_create(content_artifact_bulk)
                    self._post_save(batch)

            await sync_to_async(process_batch)()
            for declarative_content in batch:
                await self.put(declarative_content)

    def _pre_save(self, batch):
        """
        A hook plugin-writers can override to save related objects prior to content unit saving.

        This is run within the same transaction as the content unit saving.

        Args:
            batch (list of :class:`~pulpcore.plugin.stages.DeclarativeContent`): The batch of
                :class:`~pulpcore.plugin.stages.DeclarativeContent` objects to be saved.

        """
        pass

    def _post_save(self, batch):
        """
        A hook plugin-writers can override to save related objects after content unit saving.

        This is run within the same transaction as the content unit saving.

        Args:
            batch (list of :class:`~pulpcore.plugin.stages.DeclarativeContent`): The batch of
                :class:`~pulpcore.plugin.stages.DeclarativeContent` objects to be saved.

        """
        pass


class ResolveContentFutures(Stage):
    """
    This stage resolves the futures in :class:`~pulpcore.plugin.stages.DeclarativeContent`.

    Futures results are set to the found/created :class:`~pulpcore.plugin.models.Content`.

    This is useful when data downloaded from the plugin API needs to be parsed by FirstStage to
    create additional :class:`~pulpcore.plugin.stages.DeclarativeContent` objects to be send down
    the pipeline. Consider an example where content type `Foo` references additional instances of a
    different content type `Bar`. Consider this code in FirstStage::

        # Create d_content and d_artifact for a `foo_a`
        foo_a = DeclarativeContent(...)
        # Send it in the pipeline
        await self.put(foo_a)

        ...

        foo_a_content = await foo_a.resolution()  # awaits until the foo_a reaches this stage

    This creates a "looping" pattern, of sorts, where downloaded content at the end of the pipeline
    can introduce new additional to-be-downloaded content at the beginning of the pipeline.
    On the other hand, it can impose a substantial performance decrement of batching content in the
    earlier stages.
    If you want to drop a declarative content prematurely from the pipeline, use the function
    `resolve()` to unblock the coroutines awaiting the attached future and do not hand the content
    to the next stage.
    As a rule of thumb, sending more items into the pipeline first and awaiting their resolution
    later is better.
    """

    async def run(self):
        """
        The coroutine for this stage.

        Returns:
            The coroutine for this stage.
        """
        async for d_content in self.items():
            d_content.resolve()
            await self.put(d_content)


class ContentAssociation(Stage):
    """
    A Stages API stage that associates content units with `new_version`.

    This stage stores all content unit primary keys in memory before running. This is done to
    compute the units already associated but not received from `self._in_q`. These units are passed
    via `self._out_q` to the next stage as a :class:`django.db.models.query.QuerySet`.

    This stage creates a ProgressReport named 'Associating Content' that counts the number of units
    associated. Since it's a stream the total count isn't known until it's finished.

    If `mirror` was enabled, then content units may also be un-assocated (removed) from
    `new_version`. A ProgressReport named 'Un-Associating Content' is created that counts the number
    of units un-associated.

    Args:
        new_version (:class:`~pulpcore.plugin.models.RepositoryVersion`): The repo version this
            stage associates content with.
        mirror (bool): Whether or not to "mirror" the stream of DeclarativeContent - whether content
            not in the stream should be removed from the repository.
        args: unused positional arguments passed along to :class:`~pulpcore.plugin.stages.Stage`.
        kwargs: unused keyword arguments passed along to :class:`~pulpcore.plugin.stages.Stage`.
    """

    def __init__(self, new_version, mirror, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.new_version = new_version
        self.allow_delete = mirror

    async def run(self):
        """
        The coroutine for this stage.

        Returns:
            The coroutine for this stage.
        """
        async with ProgressReport(message="Associating Content", code="associating.content") as pb:
            to_delete = {
                i
                async for i in sync_to_async_iterable(
                    self.new_version.content.values_list("pk", flat=True)
                )
            }

            async for batch in self.batches():
                to_add = set()
                for d_content in batch:
                    try:
                        to_delete.remove(d_content.content.pk)
                    except KeyError:
                        to_add.add(d_content.content.pk)
                        await self.put(d_content)

                if to_add:
                    await sync_to_async(self.new_version.add_content)(
                        Content.objects.filter(pk__in=to_add)
                    )
                    await pb.aincrease_by(len(to_add))

            if self.allow_delete:
                async with ProgressReport(
                    message="Un-Associating Content", code="unassociating.content"
                ) as pb:
                    if to_delete:
                        await sync_to_async(self.new_version.remove_content)(
                            Content.objects.filter(pk__in=to_delete)
                        )
                        await pb.aincrease_by(len(to_delete))



"""
Django models related to the Status API
"""
from datetime import timedelta

from django.conf import settings
from django.db import models
from django.utils import timezone

from pulpcore.app.models import BaseModel


class ContentAppStatusManager(models.Manager):
    def online(self):
        """
        Returns a queryset of ``ContentAppStatus`` objects that are online.

        To be considered 'online', a ContentAppStatus must have a heartbeat timestamp within
        ``settings.CONTENT_APP_TTL`` from now.

        Returns:
            :class:`django.db.models.query.QuerySet`:  A query set of the ``ContentAppStatus``
                objects which are considered 'online'.
        """
        now = timezone.now()
        age_threshold = now - timedelta(seconds=settings.CONTENT_APP_TTL)

        return self.filter(last_heartbeat__gte=age_threshold)


class ContentAppStatus(BaseModel):
    """
    Represents a Content App Status

    Fields:

        name (models.TextField): The name of the content app
        last_heartbeat (models.DateTimeField): A timestamp of this worker's last heartbeat
    """

    objects = ContentAppStatusManager()

    name = models.TextField(db_index=True, unique=True)
    last_heartbeat = models.DateTimeField(auto_now=True)

    @property
    def online(self):
        """
        Whether a content app can be considered 'online'

        To be considered 'online', a content app must have a heartbeat timestamp more recent than
        the ``CONTENT_APP_TTL`` setting.

        Returns:
            bool: True if the content app is considered online, otherwise False
        """
        now = timezone.now()
        age_threshold = now - timedelta(seconds=settings.CONTENT_APP_TTL)

        return self.last_heartbeat >= age_threshold

    @property
    def missing(self):
        """
        Whether a Content App can be considered 'missing'

        To be considered 'missing', a Content App must have a timestamp older than
        ``SETTINGS.CONTENT_APP_TTL``.

        Returns:
            bool: True if the content app is considered missing, otherwise False
        """
        return not self.online

    def save_heartbeat(self):
        """
        Update the last_heartbeat field to now and save it.

        Only the last_heartbeat field will be saved. No other changes will be saved.

        Raises:
            ValueError: When the model instance has never been saved before. This method can
                only update an existing database record.
        """
        self.save(update_fields=["last_heartbeat"])



# Generated by Django 2.2.15 on 2020-09-02 04:27

import uuid

from django.db import migrations, models


def fill_in_resource_id(apps, schema_editor):
    # Fills in _resource_job_id with a random UUID, it won't matter because
    # it's only used for new ones.
    Task = apps.get_model('core', 'Task')
    for task in Task.objects.all():
        task._resource_job_id = uuid.uuid4()
        task.save()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0045_accesspolicy_permissions_allow_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='_resource_job_id',
            field=models.UUIDField(null=True),
        ),
        migrations.RunPython(fill_in_resource_id),
        migrations.AlterField(
            model_name='task',
            name='_resource_job_id',
            field=models.UUIDField()
        )
    ]



from drf_spectacular.utils import extend_schema
from rest_framework.viewsets import ViewSet

from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import AsyncOperationResponseSerializer, OrphansCleanupSerializer
from pulpcore.app.tasks import orphan_cleanup
from pulpcore.tasking.tasks import dispatch


class OrphansCleanupViewset(ViewSet):
    serializer_class = OrphansCleanupSerializer

    @extend_schema(
        description="Trigger an asynchronous orphan cleanup operation.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def cleanup(self, request):
        """
        Triggers an asynchronous orphan cleanup operation.
        """
        serializer = OrphansCleanupSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        content_pks = serializer.validated_data.get("content_hrefs", None)
        orphan_protection_time = serializer.validated_data.get("orphan_protection_time")

        task = dispatch(
            orphan_cleanup,
            kwargs={"content_pks": content_pks, "orphan_protection_time": orphan_protection_time},
        )

        return OperationPostponedResponse(task, request)



import json
import unittest
from uuid import uuid4

from pulp_smash import config
from pulp_smash.pulp3.bindings import monitor_task

from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401
from pulpcore.client.pulp_file import (
    ApiClient as FileApiClient,
    RepositoriesFileApi,
)
from pulpcore.client.pulp_file.exceptions import ApiException


class BaseLabelTestCase(unittest.TestCase):
    """Base class for label test classes."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.repo = None

    def setUp(self):
        """Create an API client."""
        self.client = FileApiClient(self.cfg.get_bindings_config())
        self.repo_api = RepositoriesFileApi(self.client)

    def _create_repo(self, labels={}):
        attrs = {"name": str(uuid4())}
        if labels:
            attrs["pulp_labels"] = labels
        self.repo = self.repo_api.create(attrs)
        self.addCleanup(self.repo_api.delete, self.repo.pulp_href)


class CRUDLabelTestCase(BaseLabelTestCase):
    """CRUD labels on repositories."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.repo = None

    def setUp(self):
        """Create an API client."""
        self.client = FileApiClient(self.cfg.get_bindings_config())
        self.repo_api = RepositoriesFileApi(self.client)

    def _create_repo(self, labels={}):
        attrs = {"name": str(uuid4())}
        if labels:
            attrs["pulp_labels"] = labels
        self.repo = self.repo_api.create(attrs)
        self.addCleanup(self.repo_api.delete, self.repo.pulp_href)

    def test_create_repo_with_labels(self):
        """Create repository with labels."""
        labels = {"maiar": "mithrandir"}
        self._create_repo(labels)
        self.assertEqual(labels, self.repo.pulp_labels)

    def test_add_repo_labels(self):
        """Update repository with labels."""
        labels = {"maiar": "mithrandir", "valar": "varda"}
        self._create_repo()

        resp = self.repo_api.partial_update(self.repo.pulp_href, {"pulp_labels": labels})
        monitor_task(resp.task)
        self.repo = self.repo_api.read(self.repo.pulp_href)
        self.assertEqual(labels, self.repo.pulp_labels)

    def test_update_repo_label(self):
        """Test updating an existing label."""
        labels = {"valar": "varda"}
        self._create_repo(labels)

        labels["valar"] = "manwe"

        resp = self.repo_api.partial_update(self.repo.pulp_href, {"pulp_labels": labels})
        monitor_task(resp.task)
        self.repo = self.repo_api.read(self.repo.pulp_href)
        self.assertEqual(labels, self.repo.pulp_labels)

    def test_unset_repo_label(self):
        """Test unsetting a repo label."""
        labels = {"maiar": "mithrandir", "valar": "varda"}
        self._create_repo(labels)

        labels.pop("valar")
        resp = self.repo_api.partial_update(self.repo.pulp_href, {"pulp_labels": labels})
        monitor_task(resp.task)
        self.repo = self.repo_api.read(self.repo.pulp_href)
        self.assertEqual(1, len(self.repo.pulp_labels))
        self.assertEqual(labels, self.repo.pulp_labels)

    def test_remove_all_repo_labels(self):
        """Test removing all labels."""
        labels = {"maiar": "mithrandir", "valar": "varda"}
        self._create_repo(labels)

        resp = self.repo_api.partial_update(self.repo.pulp_href, {"pulp_labels": {}})
        monitor_task(resp.task)
        self.repo = self.repo_api.read(self.repo.pulp_href)
        self.assertEqual(0, len(self.repo.pulp_labels))
        self.assertEqual({}, self.repo.pulp_labels)

    def test_model_partial_update(self):
        """Test that labels aren't unset accidentially with PATCH calls."""
        labels = {"maiar": "mithrandir", "valar": "varda"}
        self._create_repo(labels)

        resp = self.repo_api.partial_update(self.repo.pulp_href, {"name": str(uuid4())})
        monitor_task(resp.task)
        self.repo = self.repo_api.read(self.repo.pulp_href)
        self.assertEqual(labels, self.repo.pulp_labels)

    def test_invalid_label_type(self):
        """Test that label doesn't accept non-dicts"""
        with self.assertRaises(ApiException) as ae:
            self._create_repo("morgoth")  # str instead of dict

        self.assertEqual(400, ae.exception.status)
        self.assertTrue("pulp_labels" in json.loads(ae.exception.body))

    def test_invalid_labels(self):
        """Test that label keys and values are validated."""
        with self.assertRaises(ApiException) as ae:
            self._create_repo({"@": "maia"})

        self.assertEqual(400, ae.exception.status)
        self.assertTrue("pulp_labels" in json.loads(ae.exception.body))

        with self.assertRaises(ApiException) as ae:
            self._create_repo({"arda": "eru,illuvata"})

        self.assertEqual(400, ae.exception.status)
        self.assertTrue("pulp_labels" in json.loads(ae.exception.body))


class FilterLabelTestCase(BaseLabelTestCase):
    """CRUD labels on repositories."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.repo = None

    def setUp(self):
        """Create an API client."""
        self.client = FileApiClient(self.cfg.get_bindings_config())
        self.repo_api = RepositoriesFileApi(self.client)

    def _filter_labels(self, pulp_label_select):
        resp = self.repo_api.list(pulp_label_select=pulp_label_select)
        return resp.results

    def test_label_select(self):
        """Test removing all labels."""
        labels = {"environment": "production", "certified": "true"}
        self._create_repo(labels)
        labels = {"environment": "staging", "certified": "false"}
        self._create_repo(labels)
        labels = {}
        self._create_repo(labels)

        repos = self._filter_labels("environment=production")
        self.assertEqual(1, len(repos))

        repos = self._filter_labels("environment!=production")
        self.assertEqual(1, len(repos))

        repos = self._filter_labels("environment")
        self.assertEqual(2, len(repos))

        repos = self._filter_labels("environment~prod")
        self.assertEqual(1, len(repos))

        repos = self._filter_labels("environment=production,certified=true")
        self.assertEqual(1, len(repos))

        repos = self._filter_labels("environment=production,certified!=false")
        self.assertEqual(1, len(repos))

        repos = self._filter_labels("!environment,certified=false")
        self.assertEqual(0, len(repos))

    def test_empty_blank_filter(self):
        """Test filtering values with a blank string."""
        labels = {"environment": ""}
        self._create_repo(labels)

        repos = self._filter_labels("environment=")
        self.assertEqual(1, len(repos))

        repos = self._filter_labels("environment~")
        self.assertEqual(1, len(repos))

    def test_invalid_label_select(self):
        """Test removing all labels."""
        with self.assertRaises(ApiException) as ae:
            self._filter_labels("")
        self.assertEqual(400, ae.exception.status)

        with self.assertRaises(ApiException) as ae:
            self._filter_labels("!environment=production")
        self.assertEqual(400, ae.exception.status)

        with self.assertRaises(ApiException) as ae:
            self._filter_labels("=bad filter")
        self.assertEqual(400, ae.exception.status)



# Generated by Django 2.2.17 on 2021-01-27 18:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0055_label'),
    ]

    operations = [
        migrations.AddField(
            model_name='remote',
            name='rate_limit',
            field=models.IntegerField(null=True),
        ),
    ]



import asyncio
from contextlib import suppress
from gettext import gettext as _
from importlib import import_module
import logging
import os
import socket

from asgiref.sync import sync_to_async
from aiohttp import web

import django

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "pulpcore.app.settings")
# Until Django supports async ORM natively this is the best we can do given these parts of Pulp
# run in coroutines. We try to ensure it is safe by never passing ORM data between co-routines.
os.environ.setdefault("DJANGO_ALLOW_ASYNC_UNSAFE", "true")

django.setup()

from django.conf import settings  # noqa: E402: module level not at top of file
from django.db.utils import (  # noqa: E402: module level not at top of file
    InterfaceError,
    OperationalError,
)

from pulpcore.app.apps import pulp_plugin_configs  # noqa: E402: module level not at top of file
from pulpcore.app.models import ContentAppStatus  # noqa: E402: module level not at top of file

from .handler import Handler  # noqa: E402: module level not at top of file
from .authentication import authenticate  # noqa: E402: module level not at top of file


log = logging.getLogger(__name__)

app = web.Application(middlewares=[authenticate])

CONTENT_MODULE_NAME = "content"


async def _heartbeat():
    name = "{pid}@{hostname}".format(pid=os.getpid(), hostname=socket.gethostname())
    heartbeat_interval = settings.CONTENT_APP_TTL // 4
    i8ln_msg = _("Content App '{name}' heartbeat written, sleeping for '{interarrival}' seconds")
    msg = i8ln_msg.format(name=name, interarrival=heartbeat_interval)

    while True:

        try:
            content_app_status, created = await sync_to_async(
                ContentAppStatus.objects.get_or_create
            )(name=name)

            if not created:
                await sync_to_async(content_app_status.save_heartbeat)()

            log.debug(msg)
        except (InterfaceError, OperationalError):
            await sync_to_async(Handler._reset_db_connection)()
            i8ln_msg = _(
                "Content App '{name}' failed to write a heartbeat to the database, sleeping for "
                "'{interarrival}' seconds."
            )
            msg = i8ln_msg.format(name=name, interarrival=heartbeat_interval)
            log.info(msg)
        await asyncio.sleep(heartbeat_interval)


async def server(*args, **kwargs):
    os.chdir(settings.WORKING_DIRECTORY)

    asyncio.ensure_future(_heartbeat())
    for pulp_plugin in pulp_plugin_configs():
        if pulp_plugin.name != "pulpcore.app":
            content_module_name = "{name}.{module}".format(
                name=pulp_plugin.name, module=CONTENT_MODULE_NAME
            )
            with suppress(ModuleNotFoundError):
                import_module(content_module_name)
    app.add_routes([web.get(settings.CONTENT_PATH_PREFIX, Handler().list_distributions)])
    app.add_routes([web.get(settings.CONTENT_PATH_PREFIX + "{path:.+}", Handler().stream_content)])
    return app



import os
from datetime import datetime

from django.db import models

from pulpcore.app.models import (
    BaseModel,
    GenericRelationModel,
    MasterModel,
)
from pulpcore.app.models.repository import Repository
from pulpcore.constants import FS_EXPORT_CHOICES, FS_EXPORT_METHODS


class Export(BaseModel):
    """
    A model that represents an Export.

    Fields:

        params (models.JSONField): A set of parameters used to create the export

    Relations:

        task (models.ForeignKey): The Task that created the export
        exporter (models.ForeignKey): The Exporter that exported the resource.
    """

    params = models.JSONField(null=True)
    task = models.ForeignKey("Task", on_delete=models.PROTECT, null=True)
    exporter = models.ForeignKey("Exporter", on_delete=models.CASCADE)


class ExportedResource(GenericRelationModel):
    """
    A model to represent anything that was exported in an Export.

    Resource can be a repo version, publication, etc.

    Relations:

        export (models.ForeignKey): The Export that exported the resource.
    """

    export = models.ForeignKey(Export, related_name="exported_resources", on_delete=models.CASCADE)


class Exporter(MasterModel):
    """
    A base model that provides logic to export a set of content and keep track of Exports.

    Fields:

        name (models.TextField): The exporter unique name.
    """

    name = models.TextField(db_index=True, unique=True)


class FilesystemExport(Export):
    """
    A model that represents an export to the filesystem.
    """

    pass


class FilesystemExporter(Exporter):
    """
    A base model that provides logic to export a set of content to the filesystem.

    Fields:

        path (models.TextField): a full path where the export will go.
    """

    TYPE = "filesystem"

    path = models.TextField()
    method = models.CharField(
        choices=FS_EXPORT_CHOICES, default=FS_EXPORT_METHODS.WRITE, max_length=128
    )

    class Meta:
        default_related_name = "%(app_label)s_fs_exporter"


class PulpExport(Export):
    """
    A model that provides export-files that can be imported into other Pulp instances.

    Fields:

        tarfile (tarfile.Tarfile): a tarfile for this export to write into.
        validated_versions ([pulpcore.app.models.RepositoryVersion]): explicitly-specified versions
            to be exported (if any).
        validated_start_versions ([pulpcore.app.models.RepositoryVersion]): explicitly-specified
            starting-versions for doing an incremental export.
        validated_chunk_size (str) : requested chunk-size of the export file.
        output_file_info (models.JSONField) : JSON containing the full-path filenames and
            SHA256-checksums of all output-files generated by this export.
        toc_info (models.JSONField) : JSON containing the full-path filename and SHA256-checksum
            of the table-of-contents for this export.
    """

    tarfile = None
    validated_versions = None
    validated_start_versions = None
    validated_chunk_size = None
    output_file_info = models.JSONField(null=True)
    toc_info = models.JSONField(null=True)

    def export_tarfile_path(self):
        """
        Return the full tarfile name where the specified PulpExport should store its export
        """
        # EXPORTER-PATH/export-EXPORTID-YYYYMMDD_HHMM.tar.gz
        return os.path.normpath(
            "{}/export-{}-{}.tar.gz".format(
                self.exporter.path, str(self.pulp_id), datetime.utcnow().strftime("%Y%m%d_%H%M")
            )
        )

    class Meta:
        default_related_name = "%(app_label)s_pulp_export"


class PulpExporter(Exporter):
    """
    A model that controls creating exports that can be imported into other Pulp instances.

    Fields:

        path (models.TextField): a full path where the export will go.

    Relations:

        repositories (models.ManyToManyField): Repos to be exported.
        last_export (models.ForeignKey): The last Export from the Exporter.
    """

    TYPE = "pulp"
    path = models.TextField()
    repositories = models.ManyToManyField(Repository)
    last_export = models.ForeignKey(
        "PulpExport", related_name="last_export", on_delete=models.SET_NULL, null=True
    )

    class Meta:
        default_related_name = "%(app_label)s_pulp_exporter"



from redis import Redis
from aioredis import Redis as aRedis

from pulpcore.app.settings import settings

_conn = None
_a_conn = None


def _get_connection_from_class(redis_class):
    redis_url = settings.get("REDIS_URL")
    if redis_url is not None:
        return redis_class.from_url(redis_url)
    else:
        return redis_class(
            host=settings.REDIS_HOST,
            port=settings.REDIS_PORT,
            db=settings.REDIS_DB,
            password=settings.REDIS_PASSWORD,
            ssl=settings.REDIS_SSL,
            ssl_ca_certs=settings.REDIS_SSL_CA_CERTS,
        )


def get_redis_connection():
    global _conn

    if _conn is None:
        _conn = _get_connection_from_class(Redis)

    return _conn


def get_async_redis_connection():
    global _a_conn

    if _a_conn is None:
        _a_conn = _get_connection_from_class(aRedis)

    return _a_conn



import requests
import unittest

from urllib.parse import urljoin

from pulp_smash import config, utils
from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import gen_distribution
from pulpcore.tests.functional.api.using_plugin.constants import PULP_CONTENT_BASE_URL
from pulpcore.tests.functional.api.using_plugin.utils import gen_file_client, gen_user, del_user
from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    GroupsApi,
    GroupsUsersApi,
    ContentguardsRbacApi,
)
from pulpcore.client.pulp_file import (
    DistributionsFileApi,
    PatchedfileFileDistribution,
)

from pulpcore.tests.functional.api.using_plugin.utils import (  # noqa:F401
    set_up_module as setUpModule,
)


class RBACContentGuardTestCase(unittest.TestCase):
    """Test RBAC enabled content guard"""

    @classmethod
    def setUpClass(cls):
        """Set up test variables"""
        client = gen_file_client()  # This is admin client, following apis are for admin user
        core_client = CoreApiClient(config.get_config().get_bindings_config())
        cls.groups_api = GroupsApi(core_client)
        cls.group_users_api = GroupsUsersApi(core_client)
        cls.distro_api = DistributionsFileApi(client)
        response = monitor_task(cls.distro_api.create(gen_distribution()).task)
        cls.distro = cls.distro_api.read(response.created_resources[0])
        cls.rbac_guard_api = ContentguardsRbacApi(client)

        cls.admin = {
            "username": client.configuration.username,
            "password": client.configuration.password,
        }
        user = gen_user(model_permissions=["core.add_rbaccontentguard"])
        user["rbac_guard_api"] = ContentguardsRbacApi(user["core_api_client"])
        cls.creator_user = user
        cls.user_a = gen_user()
        cls.user_b = gen_user()
        cls.all_users = [cls.creator_user, cls.user_a, cls.user_a, cls.admin, None]

        cls.group = cls.groups_api.create({"name": utils.uuid4()})
        cls.group_users_api.create(cls.group.pulp_href, {"username": cls.user_b["username"]})
        cls.group_users_api.create(cls.group.pulp_href, {"username": cls.user_a["username"]})

        cls.url = urljoin(PULP_CONTENT_BASE_URL, f"{cls.distro.base_path}/")

    @classmethod
    def tearDownClass(cls):
        """Clean up all class variables"""
        cls.distro_api.delete(cls.distro.pulp_href)
        cls.rbac_guard_api.delete(cls.distro.content_guard)
        cls.groups_api.delete(cls.group.pulp_href)
        del_user(cls.creator_user)
        del_user(cls.user_a)
        del_user(cls.user_b)

    def test_01_all_users_access(self):
        """Sanity check that all users can access distribution with no content guard"""
        self.assert_access(self.all_users)

    def test_02_content_guard_creation(self):
        """Checks that RBAC ContentGuard can be created and assigned to a distribution"""
        guard = self.creator_user["rbac_guard_api"].create({"name": self.distro.name})
        body = PatchedfileFileDistribution(content_guard=guard.pulp_href)
        monitor_task(self.distro_api.partial_update(self.distro.pulp_href, body).task)
        RBACContentGuardTestCase.distro = self.distro_api.read(self.distro.pulp_href)
        self.assertEqual(guard.pulp_href, self.distro.content_guard)

    def test_03_only_creator_access(self):
        """Checks that now only the creator and admin user can access the distribution"""
        self.assert_access([self.creator_user, self.admin])

    def test_04_add_users(self):
        """Use the /add/ endpoint to give the users permission to access distribution"""
        body = {"usernames": (self.user_a["username"], self.user_b["username"])}
        self.creator_user["rbac_guard_api"].assign_permission(self.distro.content_guard, body)
        self.assert_access([self.creator_user, self.user_b, self.user_a, self.admin])

    def test_05_remove_users(self):
        """Use the /remove/ endpoint to remove users permission to access distribution"""
        body = {"usernames": (self.user_a["username"], self.user_b["username"])}
        self.creator_user["rbac_guard_api"].remove_permission(self.distro.content_guard, body)
        self.assert_access([self.creator_user, self.admin])

    def test_06_add_group(self):
        """Use the /add/ endpoint to add group"""
        body = {"groupnames": [self.group.name]}
        self.creator_user["rbac_guard_api"].assign_permission(self.distro.content_guard, body)
        self.assert_access([self.creator_user, self.user_b, self.user_a, self.admin])

    def test_07_remove_group(self):
        """Use the /remove/ endpoint to remove group"""
        body = {"groupnames": [self.group.name]}
        self.creator_user["rbac_guard_api"].remove_permission(self.distro.content_guard, body)
        self.assert_access([self.creator_user, self.admin])

    def assert_access(self, auth_users):
        """Helper for asserting functionality and correct permissions on the content guard"""
        for user in self.all_users:
            auth = (user["username"], user["password"]) if user else None
            r = requests.session()
            r.trust_env = False  # Don't read the .netrc file
            response = r.get(self.url, auth=auth)
            expected_status = 404 if user in auth_users else 403
            self.assertEqual(response.status_code, expected_status, f"Failed on {user=}")



from gettext import gettext as _

from django.db.models import Q
from drf_spectacular.utils import extend_schema_field
from guardian.shortcuts import get_users_with_perms, get_groups_with_perms
from rest_framework import serializers
from rest_framework.validators import UniqueValidator

from pulpcore.app import models
from pulpcore.app.serializers import (
    BaseURLField,
    DetailIdentityField,
    DetailRelatedField,
    LabelsField,
    ModelSerializer,
    RepositoryVersionRelatedField,
    validate_unknown_fields,
)
from pulpcore.app.serializers.user import GroupUserSerializer, GroupSerializer


class PublicationSerializer(ModelSerializer):
    pulp_href = DetailIdentityField(view_name_pattern=r"publications(-.*/.*)-detail")
    repository_version = RepositoryVersionRelatedField(required=False)
    repository = DetailRelatedField(
        help_text=_("A URI of the repository to be published."),
        required=False,
        label=_("Repository"),
        view_name_pattern=r"repositories(-.*/.*)?-detail",
        queryset=models.Repository.objects.all(),
    )

    def validate(self, data):
        if hasattr(self, "initial_data"):
            validate_unknown_fields(self.initial_data, self.fields)

        repository = data.pop("repository", None)  # not an actual field on publication
        repository_version = data.get("repository_version")
        if not repository and not repository_version:
            raise serializers.ValidationError(
                _("Either the 'repository' or 'repository_version' need to be specified")
            )
        elif not repository and repository_version:
            return data
        elif repository and not repository_version:
            version = repository.latest_version()
            if version:
                new_data = {"repository_version": version}
                new_data.update(data)
                return new_data
            else:
                raise serializers.ValidationError(
                    detail=_("Repository has no version available to create Publication from")
                )
        raise serializers.ValidationError(
            _(
                "Either the 'repository' or 'repository_version' need to be specified "
                "but not both."
            )
        )

    class Meta:
        abstract = True
        model = models.Publication
        fields = ModelSerializer.Meta.fields + ("repository_version", "repository")


class ContentGuardSerializer(ModelSerializer):
    pulp_href = DetailIdentityField(view_name_pattern=r"contentguards(-.*/.*)-detail")

    name = serializers.CharField(help_text=_("The unique name."))
    description = serializers.CharField(
        help_text=_("An optional description."), allow_null=True, required=False
    )

    class Meta:
        model = models.ContentGuard
        fields = ModelSerializer.Meta.fields + ("name", "description")


class RBACContentGuardSerializer(ContentGuardSerializer):

    users = serializers.SerializerMethodField()
    groups = serializers.SerializerMethodField()

    @extend_schema_field(GroupUserSerializer(many=True))
    def get_users(self, obj):
        """Finds all the users with this object's download permission."""
        users = get_users_with_perms(
            obj, with_group_users=False, only_with_perms_in=["download_rbaccontentguard"]
        )
        return GroupUserSerializer(users, many=True, context=self.context).data

    @extend_schema_field(GroupSerializer(many=True))
    def get_groups(self, obj):
        """Finds all the groups with this object's download permission."""
        groups = get_groups_with_perms(obj, attach_perms=True)
        return GroupSerializer(
            (group for group, perms in groups.items() if "download_rbaccontentguard" in perms),
            many=True,
            context=self.context,
        ).data

    class Meta:
        model = models.RBACContentGuard
        fields = ContentGuardSerializer.Meta.fields + ("users", "groups")


class RBACContentGuardPermissionSerializer(serializers.Serializer):

    usernames = serializers.ListField(default=[])
    groupnames = serializers.ListField(default=[])


class DistributionSerializer(ModelSerializer):
    """
    The Serializer for the Distribution model.

    The serializer deliberately omits the the `publication` and `repository_version` field due to
    plugins typically requiring one or the other but not both.

    To include the ``publication`` field, it is recommended plugins define the field::

      publication = DetailRelatedField(
          required=False,
          help_text=_("Publication to be served"),
          view_name_pattern=r"publications(-.*/.*)?-detail",
          queryset=models.Publication.objects.exclude(complete=False),
          allow_null=True,
      )

    To include the ``repository_version`` field, it is recommended plugins define the field::

      repository_version = RepositoryVersionRelatedField(
          required=False, help_text=_("RepositoryVersion to be served"), allow_null=True
      )

    Additionally, the serializer omits the ``remote`` field, which is used for pull-through caching
    feature and only by plugins which use publications. Plugins implementing a pull-through caching
    should define the field in their derived serializer class like this::

      remote = DetailRelatedField(
          required=False,
          help_text=_('Remote that can be used to fetch content when using pull-through caching.'),
          queryset=models.Remote.objects.all(),
          allow_null=True
      )

    """

    pulp_href = DetailIdentityField(view_name_pattern=r"distributions(-.*/.*)-detail")
    pulp_labels = LabelsField(required=False)

    base_path = serializers.CharField(
        help_text=_(
            'The base (relative) path component of the published url. Avoid paths that \
                    overlap with other distribution base paths (e.g. "foo" and "foo/bar")'
        ),
        validators=[UniqueValidator(queryset=models.Distribution.objects.all())],
    )
    base_url = BaseURLField(
        source="base_path",
        read_only=True,
        help_text=_("The URL for accessing the publication as defined by this distribution."),
    )
    content_guard = DetailRelatedField(
        required=False,
        help_text=_("An optional content-guard."),
        view_name_pattern=r"contentguards(-.*/.*)?-detail",
        queryset=models.ContentGuard.objects.all(),
        allow_null=True,
    )
    name = serializers.CharField(
        help_text=_("A unique name. Ex, `rawhide` and `stable`."),
        validators=[
            UniqueValidator(queryset=models.Distribution.objects.all()),
        ],
    )
    repository = DetailRelatedField(
        required=False,
        help_text=_("The latest RepositoryVersion for this Repository will be served."),
        view_name_pattern=r"repositories(-.*/.*)?-detail",
        queryset=models.Repository.objects.all(),
        allow_null=True,
    )

    class Meta:
        abstract = True
        model = models.Distribution
        fields = ModelSerializer.Meta.fields + (
            "base_path",
            "base_url",
            "content_guard",
            "pulp_labels",
            "name",
            "repository",
        )

    def _validate_path_overlap(self, path):
        # look for any base paths nested in path
        search = path.split("/")[0]
        q = Q(base_path=search)
        for subdir in path.split("/")[1:]:
            search = "/".join((search, subdir))
            q |= Q(base_path=search)

        # look for any base paths that nest path
        q |= Q(base_path__startswith="{}/".format(path))
        qs = models.Distribution.objects.filter(q)

        if self.instance is not None:
            qs = qs.exclude(pk=self.instance.pk)

        match = qs.first()
        if match:
            raise serializers.ValidationError(
                detail=_("Overlaps with existing distribution '{}'").format(match.name)
            )

        return path

    def validate_base_path(self, path):
        self._validate_relative_path(path)
        return self._validate_path_overlap(path)

    def validate(self, data):
        super().validate(data)

        repository_provided = data.get("repository", None) or (
            self.instance and self.instance.repository
        )
        repository_version_provided = data.get("repository_version", None) or (
            self.instance and self.instance.repository_version
        )
        publication_provided = data.get("publication", None) or (
            self.instance and self.instance.publication
        )

        if publication_provided and repository_version_provided:
            raise serializers.ValidationError(
                _(
                    "Only one of the attributes 'publication' and 'repository_version' "
                    "may be used simultaneously."
                )
            )
        elif repository_provided and repository_version_provided:
            raise serializers.ValidationError(
                _(
                    "Only one of the attributes 'repository' and 'repository_version' "
                    "may be used simultaneously."
                )
            )
        # TODO: https://pulp.plan.io/issues/8762

        # elif repository_provided and publication_provided:
        #     raise serializers.ValidationError(
        #         _(
        #             "Only one of the attributes 'repository' and 'publication' "
        #             "may be used simultaneously."
        #         )
        #     )

        return data



from gettext import gettext as _

from rest_framework import serializers

from pulpcore.app.serializers.task import ContentAppStatusSerializer, WorkerSerializer


class VersionSerializer(serializers.Serializer):
    """
    Serializer for the version information of Pulp components
    """

    component = serializers.CharField(help_text=_("Name of a versioned component of Pulp"))

    version = serializers.CharField(help_text=_("Version of the component (e.g. 3.0.0)"))


class DatabaseConnectionSerializer(serializers.Serializer):
    """
    Serializer for the database connection information
    """

    connected = serializers.BooleanField(
        help_text=_("Info about whether the app can connect to the database")
    )


class RedisConnectionSerializer(serializers.Serializer):
    """
    Serializer for information about the Redis connection
    """

    connected = serializers.BooleanField(
        help_text=_("Info about whether the app can connect to Redis")
    )


class StorageSerializer(serializers.Serializer):
    """
    Serializer for information about the storage system
    """

    total = serializers.IntegerField(min_value=0, help_text=_("Total number of bytes"))

    used = serializers.IntegerField(min_value=0, help_text=_("Number of bytes in use"))

    free = serializers.IntegerField(min_value=0, help_text=_("Number of free bytes"))


class StatusSerializer(serializers.Serializer):
    """
    Serializer for the status information of the app
    """

    versions = VersionSerializer(help_text=_("Version information of Pulp components"), many=True)

    online_workers = WorkerSerializer(
        help_text=_(
            "List of online workers known to the application. An online worker is actively "
            "heartbeating and can respond to new work"
        ),
        many=True,
    )

    online_content_apps = ContentAppStatusSerializer(
        help_text=_(
            "List of online content apps known to the application. An online content app "
            "is actively heartbeating and can serve data to clients"
        ),
        many=True,
    )

    database_connection = DatabaseConnectionSerializer(
        help_text=_("Database connection information")
    )

    redis_connection = RedisConnectionSerializer(
        required=False,
        help_text=_("Redis connection information"),
    )

    storage = StorageSerializer(required=False, help_text=_("Storage information"))



# Generated by Django 2.2.18 on 2021-03-03 10:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0058_accesspolicy_customized'),
    ]

    operations = [
        migrations.AddField(
            model_name='remote',
            name='proxy_password',
            field=models.TextField(null=True),
        ),
        migrations.AddField(
            model_name='remote',
            name='proxy_username',
            field=models.TextField(null=True),
        ),
    ]



# Generated by Django 2.2.16 on 2020-10-30 09:31

from django.db import migrations, models
import pulpcore.app.models.upload


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0048_fips_checksums'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='upload',
            name='file',
        ),
        migrations.AddField(
            model_name='uploadchunk',
            name='file',
            field=models.FileField(default=None, max_length=255, upload_to=pulpcore.app.models.upload.UploadChunk.storage_path),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-19 20:40

from django.db import migrations, transaction
from django.db.models import F


def plus_one(apps, schema_editor):
    # re: https://pulp.plan.io/issues/6147
    with transaction.atomic():
        Repository = apps.get_model('core', 'Repository')
        Repository.objects.all().update(last_version=F('last_version') + 1)


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0021_add_signing_service_foreign_key'),
    ]

    operations = [
        migrations.RunPython(plus_one),
        migrations.RenameField(
            model_name='repository',
            old_name='last_version',
            new_name='next_version',
        ),
    ]



# Generated by Django 2.2.13 on 2020-07-14 17:02

from django.db import migrations


TASK_STATEMENTS = [
    {
        "action": ["list"],
        "principal": "authenticated",
        "effect": "allow"
    }, {
        "action": ["retrieve"],
        "principal": "authenticated",
        "effect": "allow",
        "condition": "has_model_or_obj_perms:core.view_task"
    }, {
        "action": ["destroy"],
        "principal": "authenticated",
        "effect": "allow",
        "condition": "has_model_or_obj_perms:core.delete_task"
    }, {
        "action": ["update", "partial_update"],
        "principal": "authenticated",
        "effect": "allow",
        "condition": "has_model_or_obj_perms:core.change_task"
    }
]


TASK_PERMISSIONS_ASSIGNMENT = [
    {
        "function": "add_for_object_creator",
        "parameters": None,
        "permissions": ["core.view_task", "core.change_task", "core.delete_task"]
    }
]


def populate_access_policy(apps, schema_editor):
    AccessPolicy = apps.get_model('core', 'AccessPolicy')
    AccessPolicy.objects.create(
        viewset_name="TaskViewSet",
        statements=TASK_STATEMENTS,
        permissions_assignment=TASK_PERMISSIONS_ASSIGNMENT
    )


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0041_accesspolicy'),
    ]

    operations = [
        migrations.RunPython(populate_access_policy),
    ]



from concurrent.futures import ThreadPoolExecutor
from gettext import gettext as _
from logging import getLogger
import asyncio
import hashlib

from asgiref.sync import sync_to_async
from django.core.files.storage import default_storage
from django.db import transaction
from rest_framework.serializers import ValidationError

from pulpcore.app import models
from pulpcore.app.models import ProgressReport

from pulpcore.plugin.sync import sync_to_async_iterable

log = getLogger(__name__)


CHUNK_SIZE = 1024 * 1024  # 1 Mb


def delete_version(pk):
    """
    Delete a repository version by squashing its changes with the next newer version. This ensures
    that the content set for each version stays the same.

    There must be a newer version to squash into. If we deleted the latest version, the next content
    change would create a new one of the same number, which would violate the immutability
    guarantee.

    Args:
        pk (uuid): the primary key for a RepositoryVersion to delete

    Raises:
        models.RepositoryVersion.DoesNotExist: if there is not a newer version to squash into.
            TODO: something more friendly
        ValidationError: if there's one repo version
    """
    with transaction.atomic():
        try:
            version = models.RepositoryVersion.objects.get(pk=pk)
        except models.RepositoryVersion.DoesNotExist:
            log.info(_("The repository version was not found. Nothing to do."))
            return

        if version.repository.versions.complete().count() <= 1:
            raise ValidationError(
                _(
                    "Cannot delete repository version. Repositories must have at least one "
                    "repository version."
                )
            )

        log.info(
            _("Deleting and squashing version {num} of repository '{repo}'").format(
                num=version.number, repo=version.repository.name
            )
        )

        version.delete()


async def _repair_ca(content_artifact, repaired=None):
    remote_artifacts = sync_to_async_iterable(
        content_artifact.remoteartifact_set.all().select_related("remote")
    )

    if not remote_artifacts:
        log.warn(
            _("Artifact {} is unrepairable - no remote source".format(content_artifact.artifact))
        )
        return False

    async for remote_artifact in remote_artifacts:
        detail_remote = await sync_to_async(remote_artifact.remote.cast)()
        downloader = detail_remote.get_downloader(remote_artifact)
        dl_result = await downloader.run()
        if dl_result.artifact_attributes["sha256"] == content_artifact.artifact.sha256:
            with open(dl_result.path, "rb") as src:
                filename = content_artifact.artifact.file.name
                await sync_to_async(content_artifact.artifact.file.delete)(save=False)
                await sync_to_async(content_artifact.artifact.file.save)(filename, src, save=False)
            if repaired is not None:
                await repaired.aincrement()
            return True
        log.warn(_("Redownload failed from {}.").format(remote_artifact.url))

    return False


def _verify_artifact(artifact):
    try:
        # verify files digest
        hasher = hashlib.sha256()
        with artifact.file as fp:
            for chunk in fp.chunks(CHUNK_SIZE):
                hasher.update(chunk)
        return hasher.hexdigest() == artifact.sha256
    except FileNotFoundError:
        return False


async def _repair_artifacts_for_content(subset=None, verify_checksums=True):
    loop = asyncio.get_event_loop()
    pending = set()

    query_set = models.ContentArtifact.objects.exclude(artifact__isnull=True)

    if subset is not None and await sync_to_async(subset.exists)():
        query_set = query_set.filter(content__in=subset)

    async with ProgressReport(
        message="Identify missing units", code="repair.missing"
    ) as missing, ProgressReport(
        message="Identify corrupted units", code="repair.corrupted"
    ) as corrupted, ProgressReport(
        message="Repair corrupted units", code="repair.repaired"
    ) as repaired:

        with ThreadPoolExecutor(max_workers=2) as checksum_executor:
            async for content_artifact in sync_to_async_iterable(
                query_set.select_related("artifact").iterator()
            ):
                artifact = content_artifact.artifact

                valid = await loop.run_in_executor(None, default_storage.exists, artifact.file.name)
                if not valid:
                    await missing.aincrement()
                    log.warn(_("Missing file for {}").format(artifact))
                elif verify_checksums:
                    # default ThreadPoolExecutor uses num cores x 5 threads. Since we're doing
                    # such long and sequential reads, using too many threads might hurt more
                    # than help (on HDDs, maybe not on SSDs) by making the disk access pattern
                    # more random. Put it in a separate executor with limited threads.
                    # Should stay in (an) executor so that at least it doesn't completely block
                    # downloads.
                    valid = await loop.run_in_executor(
                        checksum_executor, _verify_artifact, artifact
                    )
                    if not valid:
                        await corrupted.aincrement()
                        log.warn(_("Digest mismatch for {}").format(artifact))

                if not valid:
                    if len(pending) >= 5:  # Limit the number of concurrent repair tasks
                        done, pending = await asyncio.wait(
                            pending, return_when=asyncio.FIRST_COMPLETED
                        )
                        await asyncio.gather(*done)  # Clean up tasks
                    pending.add(asyncio.ensure_future(_repair_ca(content_artifact, repaired)))
        await asyncio.gather(*pending)


def repair_version(repository_version_pk, verify_checksums):
    """
    Repair the artifacts associated with this repository version.

    Artifact files that have suffered from bit rot, were altered or have gone missing will be
    attempted to be refetched fron an associated upstream.

    Args:
        repository_version_pk (uuid): the primary key for a RepositoryVersion to delete
        verify_checksums (bool): whether to calculate and compare checksums for all artifacts
    """

    version = models.RepositoryVersion.objects.get(pk=repository_version_pk)

    log.info(
        _("Repairing version {num} of repository '{repo}'").format(
            num=version.number, repo=version.repository.name
        ),
    )

    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        _repair_artifacts_for_content(subset=version.content, verify_checksums=verify_checksums)
    )


def repair_all_artifacts(verify_checksums):
    """
    Repair all artifacts, globally.

    Artifact files that have suffered from bit rot, were altered or have gone missing will be
    attempted to be refetched fron an associated upstream.

    Args:
        verify_checksums (bool): whether to calculate and compare checksums for all artifacts
    """
    log.info(_("Repairing artifacts.'"))

    loop = asyncio.get_event_loop()
    loop.run_until_complete(_repair_artifacts_for_content(verify_checksums=verify_checksums))


def add_and_remove(repository_pk, add_content_units, remove_content_units, base_version_pk=None):
    """
    Create a new repository version by adding and then removing content units.

    Args:
        repository_pk (int): The primary key for a Repository for which a new Repository Version
            should be created.
        add_content_units (list): List of PKs for :class:`~pulpcore.app.models.Content` that
            should be added to the previous Repository Version for this Repository.
        remove_content_units (list): List of PKs for:class:`~pulpcore.app.models.Content` that
            should be removed from the previous Repository Version for this Repository.
        base_version_pk (int): the primary key for a RepositoryVersion whose content will be used
            as the initial set of content for our new RepositoryVersion
    """
    repository = models.Repository.objects.get(pk=repository_pk).cast()

    if base_version_pk:
        base_version = models.RepositoryVersion.objects.get(pk=base_version_pk)
    else:
        base_version = None

    if "*" in remove_content_units:
        latest = repository.latest_version()
        if latest:
            remove_content_units = latest.content.values_list("pk", flat=True)
        else:
            remove_content_units = []

    with repository.new_version(base_version=base_version) as new_version:
        new_version.remove_content(models.Content.objects.filter(pk__in=remove_content_units))
        new_version.add_content(models.Content.objects.filter(pk__in=add_content_units))



from gettext import gettext as _

from django.db import connection
from django.conf import settings
from django.core.management import BaseCommand
from django.db.models import Q
from django.utils.encoding import force_bytes, force_str

import cryptography

from pulpcore.app.models import Remote


class Command(BaseCommand):
    """
    Django management command for repairing incorrectly migrated remote data.
    """

    help = _(
        "Repairs issue #2327. A small number of configuration settings may have been "
        "corrupted during an upgrade from a previous version of Pulp to a Pulp version "
        "between 3.15-3.18, resulting in trouble when syncing or viewing certain remotes. "
        "This script repairs the data (which was not lost)."
    )

    def add_arguments(self, parser):
        """Set up arguments."""
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help=_("Don't modify anything, just collect results on how many Remotes are impacted."),
        )

    def handle(self, *args, **options):

        dry_run = options["dry_run"]
        fields = ("username", "password", "proxy_username", "proxy_password", "client_key")

        with open(settings.DB_ENCRYPTION_KEY, "rb") as key_file:
            fernet = cryptography.fernet.Fernet(key_file.read())

        possibly_affected_remotes = (
            Q(username__isnull=False)
            | Q(password__isnull=False)
            | Q(proxy_username__isnull=False)
            | Q(proxy_password__isnull=False)
            | Q(client_key__isnull=False)
        )

        number_unencrypted = 0
        number_multi_encrypted = 0

        for remote_pk in Remote.objects.filter(possibly_affected_remotes).values_list(
            "pk", flat=True
        ):
            try:
                remote = Remote.objects.get(pk=remote_pk)
                # if we can get the remote successfully, it is either OK or the fields are
                # encrypted more than once
            except cryptography.fernet.InvalidToken:
                # If decryption fails then it probably hasn't been encrypted yet
                # get the raw column value, avoiding any Django field handling
                with connection.cursor() as cursor:
                    cursor.execute(
                        "SELECT username, password, proxy_username, proxy_password, client_key "
                        "FROM core_remote WHERE pulp_id = %s",
                        [str(remote_pk)],
                    )
                    row = cursor.fetchone()

                field_values = {}

                for field, value in zip(fields, row):
                    field_values[field] = value

                if not dry_run:
                    Remote.objects.filter(pk=remote_pk).update(**field_values)
                number_unencrypted += 1
            else:
                times_decrypted = 0
                keep_trying = True
                needs_update = False

                while keep_trying:
                    for field in fields:
                        field_value = getattr(remote, field)  # value gets decrypted once on access
                        if not field_value:
                            continue

                        try:
                            # try to decrypt it again
                            field_value = force_str(fernet.decrypt(force_bytes(field_value)))
                            # it was decrypted successfully again time, so it was probably
                            # encrypted multiple times over. lets re-set the value with the
                            # newly decrypted value
                            setattr(remote, field, field_value)
                            needs_update = True
                        except cryptography.fernet.InvalidToken:
                            # couldn't be decrypted again, stop here
                            keep_trying = False

                    times_decrypted += 1

                if needs_update:
                    if not dry_run:
                        remote.save()
                    number_multi_encrypted += 1

        if dry_run:
            print("Remotes with un-encrypted fields: {}".format(number_unencrypted))
            print("Remotes encrypted multiple times: {}".format(number_multi_encrypted))
        else:
            if not number_unencrypted and not number_multi_encrypted:
                print("Finished. (OK)")
            else:
                print(
                    "Finished. ({} remotes fixed)".format(
                        number_unencrypted + number_multi_encrypted
                    )
                )



# Generated by Django 2.2.9 on 2020-01-30 09:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0020_change_publishedartifact_constraints'),
    ]

    operations = [
        migrations.CreateModel(
            name='AsciiArmoredDetachedSigningService',
            fields=[
                ('signingservice_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.SigningService')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.signingservice',),
        ),
    ]



# Generated by Django 2.2.11 on 2020-05-22 18:31

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0031_import_export_validate_params'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='pulpexport',
            name='filename',
        ),
        migrations.RemoveField(
            model_name='pulpexport',
            name='sha256',
        ),
        migrations.AddField(
            model_name='pulpexport',
            name='output_file_info',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
    ]



from import_export import resources


class QueryModelResource(resources.ModelResource):
    """
    A ModelResource that knows the RepositoryVersion to use to filter its query

    QueryModelResource has-a repository-version that can be used to limit its export, and a
    queryset that is derived from that repository-version.

    A plugin-writer will subclass their ModelResources from QueryModelResource,
    and use it to define the limiting query

    Attributes:

        repo_version (models.RepositoryVersion): The RepositoryVersion whose content we would like
            to export
        queryset (django.db.models.query.QuerySet): filtering queryset for this resource
            (driven by repo_version)
    """

    def set_up_queryset(self):
        return None

    def __init__(self, repo_version=None):
        self.repo_version = repo_version
        if repo_version:
            self.queryset = self.set_up_queryset()

    class Meta:
        exclude = ("pulp_id", "pulp_created", "pulp_last_updated")


class BaseContentResource(QueryModelResource):
    """
    A QueryModelResource that knows how to fill in the 'upstream_id' export-field

    BaseContentResource knows to de/hydrate upstream_id with the content-being-exported's pulp_id.

    All Content-based resources being import/exported should subclass from this class.
    """

    # An optional mapping that maps Content to Repositories. Useful when Content is sometimes not
    # tied directly to a Repository but rather to a subrepo. Formatting:
    #
    #     {"<repo name>": ["<content upstream_id>", "..."]}
    #
    content_mapping = None

    class Meta:
        exclude = QueryModelResource.Meta.exclude + ("_artifacts", "content", "content_ptr")

    def dehydrate_upstream_id(self, content):
        return str(content.pulp_id)



from gettext import gettext as _
from logging import getLogger
import re
import traceback
from urllib.parse import urljoin

from django.core.validators import URLValidator
from django.db import transaction
from drf_queryfields.mixins import QueryFieldsMixin
from rest_framework import serializers
from rest_framework_nested.relations import (
    NestedHyperlinkedIdentityField,
    NestedHyperlinkedRelatedField,
)

from pulpcore.app.models import Label, Task, TaskGroup
from pulpcore.app.util import (
    get_view_name_for_model,
    get_viewset_for_model,
    get_request_without_query_params,
)


log = getLogger(__name__)


def validate_unknown_fields(initial_data, defined_fields):
    """
    This will raise a `ValidationError` if a serializer is passed fields that are unknown.
    The `csrfmiddlewaretoken` field is silently ignored.
    """
    ignored_fields = {"csrfmiddlewaretoken"}
    unknown_fields = set(initial_data) - set(defined_fields) - ignored_fields
    if unknown_fields:
        unknown_fields = {field: _("Unexpected field") for field in unknown_fields}
        raise serializers.ValidationError(unknown_fields)


class ValidateFieldsMixin:
    """A mixin for validating unknown serializers' fields."""

    def validate(self, data):
        if hasattr(self, "initial_data"):
            validate_unknown_fields(self.initial_data, self.fields)

        data = super().validate(data)
        return data


class ModelSerializer(
    ValidateFieldsMixin, QueryFieldsMixin, serializers.HyperlinkedModelSerializer
):
    """Base serializer for use with :class:`pulpcore.app.models.Model`

    This ensures that all Serializers provide values for the 'pulp_href` field.

    The class provides a default for the ``ref_name`` attribute in the
    ModelSerializers's ``Meta`` class. This ensures that the OpenAPI definitions
    of plugins are namespaced properly.

    """

    # default is 'fields!' which doesn't work in the bindings for some langs
    exclude_arg_name = "exclude_fields"

    class Meta:
        fields = ("pulp_href", "pulp_created")

    pulp_created = serializers.DateTimeField(help_text=_("Timestamp of creation."), read_only=True)

    def _validate_relative_path(self, path):
        """
        Validate a relative path (eg from a url) to ensure it forms a valid url and does not begin
        or end with slashes nor contain spaces

        Args:
            path (str): A relative path to validate

        Returns:
            str: the validated path

        Raises:
            django.core.exceptions.ValidationError: if the relative path is invalid

        """
        # in order to use django's URLValidator we need to construct a full url
        base = "http://localhost"  # use a scheme/hostname we know are valid

        if " " in path:
            raise serializers.ValidationError(detail=_("Relative path cannot contain spaces."))

        validate = URLValidator()
        validate(urljoin(base, path))

        if path != path.strip("/"):
            raise serializers.ValidationError(
                detail=_("Relative path cannot begin or end with slashes.")
            )

        return path

    def __init_subclass__(cls, **kwargs):
        """Set default attributes in subclasses.

        Sets the default for the ``ref_name`` attribute for a ModelSerializers's
        ``Meta`` class.

        If the ``Meta.ref_name`` attribute is not yet defined, set it according
        to the best practice established within Pulp: ``<app label>.<model class
        name>``. ``app_label`` is used to create a per plugin namespace.

        Serializers in pulpcore (``app_label`` is 'core') will not be
        namespaced, i.e. ref_name is not set in this case.

        The ``ref_name`` default value is computed using ``Meta.model``. If that
        is not defined (because the class must be subclassed to be useful),
        `ref_name` is not set.

        """
        super().__init_subclass__(**kwargs)
        meta = cls.Meta
        try:
            if not hasattr(meta, "ref_name"):
                plugin_namespace = meta.model._meta.app_label
                if plugin_namespace != "core":
                    meta.ref_name = f"{plugin_namespace}.{meta.model.__name__}"
        except AttributeError:
            pass

    def _update_labels(self, instance, labels):
        """
        Update the labels for a Model instance.

        Args:
            instance (pulpcore.app.models.BaseModel): instance with labels to update
            labels (list): labels to set for the instance
        """
        instance.pulp_labels.exclude(key__in=labels.keys()).delete()

        for key, value in labels.items():
            label = instance.pulp_labels.filter(key=key).first()
            try:
                label = instance.pulp_labels.get(key=key)
                if label.value != value:
                    instance.pulp_labels.filter(key=key).update(value=value)
            except Label.DoesNotExist:
                instance.pulp_labels.create(key=key, value=value)

    def create(self, validated_data):
        """
        Created the resource from validated_data.

        Args:
            validated_data (dict): Validated data to create instance

        Returns:
            instance: The created of resource
        """
        labels = validated_data.pop("pulp_labels", {})
        with transaction.atomic():
            instance = super().create(validated_data)
            self._update_labels(instance, labels)
        return instance

    def update(self, instance, validated_data):
        """
        Update the resource from validated_data.

        Args:
            validated_data (dict): Validated data to update instance

        Returns:
            instance: The updated instance of resource
        """
        labels = validated_data.pop("pulp_labels", None)
        with transaction.atomic():
            instance = super().update(instance, validated_data)
            if labels is not None:
                self._update_labels(instance, labels)
        return instance


class _MatchingRegexViewName(object):
    """This is a helper class to help defining object matching rules for master-detail.

    If you can be specific, please specify the `view_name`, but if you cannot, this allows
    you to specify a regular expression like .e.g. `r"repositories(-.*/.*)?-detail"` to
    identify whether the provided resources viewn name belongs to any repository type.
    """

    __slots__ = ("pattern",)

    def __init__(self, pattern):
        self.pattern = pattern

    def __repr__(self):
        return f'{self.__class__.__name__}(r"{self.pattern}")'

    def __eq__(self, other):
        return re.fullmatch(self.pattern, other) is not None


class _DetailFieldMixin:
    """Mixin class containing code common to DetailIdentityField and DetailRelatedField"""

    def __init__(self, view_name=None, view_name_pattern=None, **kwargs):
        if view_name is None:
            # set view name to prevent a DRF assertion that view_name is not None
            # Anything that accesses self.view_name after __init__
            # needs to have it set before being called. Unfortunately, a model instance
            # is required to derive this value, so we can't make a view_name property.
            if view_name_pattern:
                view_name = _MatchingRegexViewName(view_name_pattern)
            else:
                log.warn(
                    _(
                        "Please provide either 'view_name' or 'view_name_pattern' for {} on {}."
                    ).format(self.__class__.__name__, traceback.extract_stack()[-4][2])
                )
                view_name = _MatchingRegexViewName(r".*")
        super().__init__(view_name, **kwargs)

    def _view_name(self, obj):
        # this is probably memoizeable based on the model class if we want to get cachey
        try:
            obj = obj.cast()
        except AttributeError:
            # The normal message that comes up here is unhelpful, so do like other DRF
            # fails do and be a little more helpful in the exception message.
            msg = (
                'Expected a detail model instance, not {}. Do you need to add "many=True" to '
                "this field definition in its serializer?"
            ).format(type(obj))
            raise ValueError(msg)
        return get_view_name_for_model(obj, "detail")

    def get_url(self, obj, view_name, request, *args, **kwargs):
        # ignore the passed in view name and return the url to the cast unit, not the generic unit
        request = None
        view_name = self._view_name(obj)
        return super().get_url(obj, view_name, request, *args, **kwargs)


class IdentityField(serializers.HyperlinkedIdentityField):
    """IdentityField for use in the pulp_href field of non-Master/Detail Serializers.

    The get_url method is overriden so relative URLs are returned.
    """

    def get_url(self, obj, view_name, request, *args, **kwargs):
        # ignore the passed in view name and return the url to the cast unit, not the generic unit
        request = None
        return super().get_url(obj, view_name, request, *args, **kwargs)


class RelatedField(serializers.HyperlinkedRelatedField):
    """RelatedField when relating to non-Master/Detail models

    When using this field on a serializer, it will serialize the related resource as a relative URL.
    """

    def get_url(self, obj, view_name, request, *args, **kwargs):
        # ignore the passed in view name and return the url to the cast unit, not the generic unit
        request = None
        return super().get_url(obj, view_name, request, *args, **kwargs)


class RelatedResourceField(RelatedField):
    """RelatedResourceField when relating a Resource object models.

    This field should be used to relate a list of non-homogeneous resources. e.g.:
    CreatedResource and ExportedResource models that store relationships to arbitrary
    resources.

    Specific implementation requires the model to be defined in the Meta:.
    """

    def to_representation(self, data):
        # If the content object was deleted
        if data.content_object is None:
            return None
        try:
            if not data.content_object.complete:
                return None
        except AttributeError:
            pass

        # query parameters can be ignored because we are looking just for 'pulp_href'; still,
        # we need to use the request object due to contextual references required by some
        # serializers
        request = get_request_without_query_params(self.context)

        viewset = get_viewset_for_model(data.content_object)
        serializer = viewset.serializer_class(data.content_object, context={"request": request})
        return serializer.data.get("pulp_href")


class DetailIdentityField(_DetailFieldMixin, serializers.HyperlinkedIdentityField):
    """IdentityField for use in the pulp_href field of Master/Detail Serializers

    When using this field on a Serializer, it will automatically cast objects to their Detail type
    base on the Serializer's Model before generating URLs for them.

    Subclasses must indicate the Master model they represent by declaring a queryset
    in their class body, usually <MasterModelImplementation>.objects.all().
    """


class DetailRelatedField(_DetailFieldMixin, serializers.HyperlinkedRelatedField):
    """RelatedField for use when relating to Master/Detail models

    When using this field on a Serializer, relate it to the Master model in a
    Master/Detail relationship, and it will automatically cast objects to their Detail type
    before generating URLs for them.

    Subclasses must indicate the Master model they represent by declaring a queryset
    in their class body, usually <MasterModelImplementation>.objects.all().
    """

    def get_object(self, *args, **kwargs):
        # return the cast object, not the generic contentunit
        return super().get_object(*args, **kwargs).cast()

    def use_pk_only_optimization(self):
        """
        If the lookup field is `pk`, DRF substitutes a PKOnlyObject as an optimization. This
        optimization breaks with Detail fields like this one which need access to their Meta
        class to get the relevant `view_name`.
        """
        return False


class NestedIdentityField(NestedHyperlinkedIdentityField):
    """NestedIdentityField for use with nested resources.

    When using this field in a serializer, it serializes the resource as a relative URL.
    """

    def get_url(self, obj, view_name, request, *args, **kwargs):
        # ignore the passed in view name and return the url to the cast unit, not the generic unit
        request = None
        return super().get_url(obj, view_name, request, *args, **kwargs)


class NestedRelatedField(NestedHyperlinkedRelatedField):
    """NestedRelatedField for use when relating to nested resources.

    When using this field in a serializer, it serializes the related resource as a relative URL.
    """

    def get_url(self, obj, view_name, request, *args, **kwargs):
        # ignore the passed in view name and return the url to the cast unit, not the generic unit
        request = None
        return super().get_url(obj, view_name, request, *args, **kwargs)


class AsyncOperationResponseSerializer(serializers.Serializer):
    """
    Serializer for asynchronous operations.
    """

    task = RelatedField(
        required=True,
        help_text=_("The href of the task."),
        queryset=Task.objects,
        view_name="tasks-detail",
        allow_null=False,
    )


class TaskGroupOperationResponseSerializer(serializers.Serializer):
    """
    Serializer for asynchronous operations that return a task group.
    """

    task_group = RelatedField(
        required=True,
        help_text=_("The href of the task group."),
        queryset=TaskGroup.objects,
        view_name="task-groups-detail",
        allow_null=False,
    )



# Generated by Django 2.2.13 on 2020-07-01 21:29

from django.contrib.auth import get_user_model
from django.db import migrations


def allow_admin_as_staff(apps, schema_editor):
    user_model = get_user_model()
    try:
        admin_user = user_model.objects.get(username='admin')
    except user_model.DoesNotExist:
        pass
    else:
        admin_user.is_staff = True
        admin_user.save()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0039_change_download_concurrency'),
    ]

    operations = [
        migrations.RunPython(allow_admin_as_staff),
    ]



# Generated by Django 3.2.5 on 2021-08-04 17:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0071_filesystemexport_filesystemexporter'),
    ]

    operations = [
        migrations.AddField(
            model_name='filesystemexporter',
            name='method',
            field=models.CharField(choices=[('write', 'Export by writing'), ('hardlink', 'Export by hardlinking'), ('symlink', 'Export by symlinking')], default='write', max_length=128),
        ),
    ]



from drf_spectacular.utils import extend_schema
from rest_framework.views import APIView

from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import AsyncOperationResponseSerializer, RepairSerializer
from pulpcore.app.tasks import repair_all_artifacts
from pulpcore.tasking.tasks import dispatch


class RepairView(APIView):
    @extend_schema(
        description=(
            "Trigger an asynchronous task that checks for missing "
            "or corrupted artifacts, and attempts to redownload them."
        ),
        summary="Repair Artifact Storage",
        request=RepairSerializer,
        responses={202: AsyncOperationResponseSerializer},
    )
    def post(self, request):
        """
        Repair artifacts.
        """
        serializer = RepairSerializer(data=request.data)
        serializer.is_valid()

        verify_checksums = serializer.validated_data["verify_checksums"]

        task = dispatch(repair_all_artifacts, args=[verify_checksums])

        return OperationPostponedResponse(task, request)



"""
Django models related to progress reporting
"""
import datetime
import logging
from asgiref.sync import sync_to_async
from asyncio import CancelledError
from gettext import gettext as _

from django.db import models
from django.utils import timezone

from pulpcore.app.models import BaseModel, Task
from pulpcore.constants import TASK_CHOICES, TASK_STATES

_logger = logging.getLogger(__name__)

# number of ms between save() calls when _using_context_manager is set
BATCH_INTERVAL = 2000


class ProgressReport(BaseModel):
    """
    A model for all progress reporting.

    All progress reports have a message, state, and are related to a Task.

    Plugin writers should create these objects to show progress reporting of a single step or
    aspect of work which has a name and state. For example:

        >>> ProgressReport(
        >>>     message='Publishing files', code='publish', total=23
        >>> )  # default: state='waiting' and done=0

        >>> ProgressReport(
        >>>     message='Publishing files', code='publish', total=23, state='running'
        >>> )  # specify the state
        >>> ProgressReport(
        >>>     message='Publishing files', code='publish', total=23, done=16
        >>> )  # already completed 16

    Update the state to COMPLETED and save it:

        >>> progress_bar = ProgressReport(
        >>>     message='Publishing files', code='publish', total=23, state='running'
        >>> )
        >>> progress_bar.state = 'completed'
        >>> progress_bar.save()

    The ProgressReport() is a context manager that provides automatic state transitions and saving
    for the RUNNING CANCELED COMPLETED and FAILED states. The increment() method can be called in
    the loop as work is completed. When ProgressReport() is used as a context manager progress
    reporting is rate limited to every 500 milliseconds.
    Use it as follows:

        >>> progress_bar = ProgressReport(
        >>>     message='Publishing files', code='publish', total=len(files_iterator)
        >>> )
        >>> progress_bar.save()
        >>> with progress_bar:
        >>>     # progress_bar saved as 'running'
        >>>     for file in files_iterator:
        >>>         handle(file)
        >>>         progress_bar.increment()  # increments and saves
        >>> # progress_bar is saved as 'completed' if no exception or 'failed' otherwise

    A convenience method called iter() allows you to avoid calling increment() directly:

        >>> progress_bar = ProgressReport(
        >>>     message='Publishing files', code='publish', total=len(files_iterator)
        >>> )
        >>> progress_bar.save()
        >>> with progress_bar:
        >>>     for file in progress_bar.iter(files_iterator):
        >>>         handle(file)

    You can also use this short form which handles all necessary save() calls:

        >>> data = dict(message='Publishing files', code='publish', total=len(files_iterator))
        >>> with ProgressReport(**data) as pb:
        >>>     for file in pb.iter(files_iterator):
        >>>         handle(file)

    ProgressReport objects are associated with a Task and auto-discover and populate the task id
    when saved.

    When using threads to update a ProgressReport in parallel, it is recommended that all threads
    share the same in-memory instance. Django does not synchronize in-memory model instances, so
    multiple instances of a specific ProgressReport will diverge as they are written to from
    separate threads.

    Fields:

        message (models.TextField): short message for the progress update, typically
            shown to the user. (required)
        code (models.TextField): identifies the type of progress report
        state (models.TextField): The state of the progress update. Defaults to `WAITING`. This
            field uses a limited set of choices of field states. See `STATES` for possible states.
        total: (models.IntegerField) The total count of items to be handled
        done (models.IntegerField): The count of items already processed. Defaults to 0.
        suffix (models.TextField): Customizable suffix rendered with the progress report
            See `the docs <https://github.com/verigak/progress>`_. for more info.

    Relations:

        task: The task associated with this progress report. If left unset when save() is called
            it will be set to the current task_id.
    """

    message = models.TextField()
    code = models.CharField(max_length=36)
    state = models.TextField(choices=TASK_CHOICES, default=TASK_STATES.WAITING)

    total = models.IntegerField(null=True)
    done = models.IntegerField(default=0)

    task = models.ForeignKey("Task", related_name="progress_reports", on_delete=models.CASCADE)

    suffix = models.TextField(null=True)

    _using_context_manager = False
    _last_save_time = None

    def save(self, *args, **kwargs):
        """
        Auto-set the task_id if running inside a task

        If the task_id is already set it will not be updated. If it is unset and this is running
        inside of a task it will be auto-set prior to saving.

        args (list): positional arguments to be passed on to the real save
        kwargs (dict): keyword arguments to be passed on to the real save
        """
        now = timezone.now()

        if not self.task_id:
            self.task = Task.current()

        if self._using_context_manager and self._last_save_time:
            if now - self._last_save_time >= datetime.timedelta(milliseconds=BATCH_INTERVAL):
                super().save(*args, **kwargs)
                self._last_save_time = now
        else:
            super().save(*args, **kwargs)
            self._last_save_time = now

    asave = sync_to_async(save)

    def __enter__(self):
        """
        Saves the progress report state as RUNNING
        """
        self.state = TASK_STATES.RUNNING
        self.save()

        # Save needs occurs immediately so it is called before _using_context_manager is set
        self._using_context_manager = True
        return self

    async def __aenter__(self):
        """
        Async implementation of __enter__
        """
        self.state = TASK_STATES.RUNNING
        await self.asave()

        # Save needs occurs immediately so it is called before _using_context_manager is set
        self._using_context_manager = True
        return self

    def __exit__(self, type, value, traceback):
        """
        Update the progress report state to COMPLETED, CANCELED, or FAILED.

        If an exception occurs the progress report state is saved as:
        - CANCELED if the exception is `asyncio.CancelledError`
        - FAILED otherwise.

        The exception is not suppressed. If the context manager exited without
        exception the progress report state is saved as COMPLETED.

        See the context manager documentation for more info on __exit__ parameters
        """
        self._using_context_manager = False
        if type is None:
            self.state = TASK_STATES.COMPLETED
        elif type is CancelledError:
            self.state = TASK_STATES.CANCELED
        else:
            self.state = TASK_STATES.FAILED
        self.save()

    async def __aexit__(self, type, value, traceback):
        """
        Async implementation of __exit__

        Update the progress report state to COMPLETED, CANCELED, or FAILED.

        If an exception occurs the progress report state is saved as:
        - CANCELED if the exception is `asyncio.CancelledError`
        - FAILED otherwise.

        The exception is not suppressed. If the context manager exited without
        exception the progress report state is saved as COMPLETED.

        See the context manager documentation for more info on __exit__ parameters
        """
        self._using_context_manager = False
        if type is None:
            self.state = TASK_STATES.COMPLETED
        elif type is CancelledError:
            self.state = TASK_STATES.CANCELED
        else:
            self.state = TASK_STATES.FAILED
        await self.asave()

    def increment(self):
        """
        Increment done count and save the progress report.

        This will increment and save the self.done attribute which is useful to put into a loop
        processing items.
        """
        self.increase_by(1)

    aincrement = sync_to_async(increment)

    def increase_by(self, count):
        """
        Increase the done count and save the progress report.

        This will increment and save the self.done attribute which is useful to put into a loop
        processing items.
        """
        self.done += count
        if self.total:
            if self.done > self.total:
                _logger.warning(_("Too many items processed for ProgressReport %s") % self.message)
        self.save()

    aincrease_by = sync_to_async(increase_by)

    def iter(self, iter):
        """
        Iterate and automatically call increment().

            >>> progress_bar = ProgressReport(message='Publishing files', code='publish', total=23)
            >>> progress_bar.save()
            >>> for file in progress_bar.iter(files_iterator):
            >>>     handle(file)

        Args:
            iter (iterator): The iterator to loop through while incrementing

        Returns:
            generator of ``iter`` argument items
        """
        for x in iter:
            yield x
            self.increment()


class GroupProgressReport(BaseModel):
    """
    A model for all progress reporting in a Task Group.

    All progress reports have a message, code and are related to a Task Group.

    Once a Task Group is created, plugin writers should create these objects ahead.
    For example:

        >>> task_group = TaskGroup(description="Migration Sub-tasks")
        >>> task_group.save()
        >>> group_pr = GroupProgressReport(
        >>>     message="Repo migration",
        >>>     code="create.repo_version",
        >>>     task_group=task_group).save()

    Taks that will be executing certain work, and is part of a TaskGroup, will look for
    the Task group it belongs to and find appropriate progress report by its code and will
    update it accordingly.

    For example:

        >>> task_group = TaskGroup.current()
        >>> progress_repo = task_group.group_progress_reports.filter(code='create.repo_version')
        >>> progress_repo.update(done=F('done') + 1)

    To avoid race conditions/cache invalidation issues, this pattern needs to be used so that
    operations are performed directly inside the database:

    .update(done=F('done') + 1)

    See: https://docs.djangoproject.com/en/3.0/ref/models/expressions/#f-expressions
    Important: F() objects assigned to model fields persist after saving the model instance and
    will be applied on each save(). Do not use save() and use update() instead, otherwise
    refresh_from_db() should be called after each save()

    Fields:

        message (models.TextField): short message for the progress update, typically
            shown to the user. (required)
        code (models.TextField): identifies the type of progress report
        total: (models.IntegerField) The total count of items to be handled
        done (models.IntegerField): The count of items already processed. Defaults to 0.
        suffix (models.TextField): Customizable suffix rendered with the progress report
            See `the docs <https://github.com/verigak/progress>`_. for more info.

    Relations:

        task_group: The task group associated with this group progress report.

    """

    message = models.TextField()
    code = models.CharField(max_length=36)

    total = models.IntegerField(default=0)
    done = models.IntegerField(default=0)

    task_group = models.ForeignKey(
        "TaskGroup", related_name="group_progress_reports", on_delete=models.CASCADE
    )

    suffix = models.TextField(null=True)

    class Meta:
        unique_together = ("code", "task_group")
